"""
Walker interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.graph.node import node
from jaseci.jac.interpreter.interp import interp
from jaseci.jac.jac_set import jac_set
from jaseci.jac.machine.jac_scope import jac_scope
from jaseci.jac.ir.jac_code import jac_ir_to_ast
from jaseci.utils.id_list import id_list


class walker_interp(interp):
    """Jac interpreter mixin for objects that will execute Jac code"""

    # Walker only executes statements, sentinels handle attr_stmts

    def run_walker(self, jac_ast):
        """
        walker: KW_WALKER NAME namespaces? walker_block;
        """
        kid = self.set_cur_ast(jac_ast)
        if jac_ast.name == "walker_block":  # used in jac tests
            self.scope_and_run(jac_ast, self.run_walker_block)
        else:
            self.scope_and_run(kid[-1], self.run_walker_block)

    def run_walker_block(self, jac_ast):
        """
        walker_block:
            LBRACE attr_stmt* walk_entry_block? (
                statement
                | walk_activity_block
            )* walk_exit_block? RBRACE;
        """
        kid = self.set_cur_ast(jac_ast)
        if self.current_step == 0:
            for i in kid:
                if i.name == "attr_stmt":
                    self.run_attr_stmt(jac_ast=i, obj=self)

        archs = self.get_arch_for(self.current_node).arch_with_supers()
        act_list = id_list(self)
        for i in archs:
            act_list += i.entry_action_ids
        self.auto_trigger_node_actions(nd=self.current_node, act_list=act_list)

        for i in kid:
            if i.name == "walk_entry_block":
                self.run_walk_entry_block(i)
            if i.name == "statement":
                self.run_statement(i)
            if i.name == "walk_activity_block":
                self.run_walk_activity_block(i)
            if i.name == "walk_exit_block":
                self.run_walk_exit_block(i)

        # self.trigger_activity_actions()
        archs = self.get_arch_for(self.current_node).arch_with_supers()
        act_list = id_list(self)
        for i in archs:
            act_list += i.exit_action_ids
        self.auto_trigger_node_actions(nd=self.current_node, act_list=act_list)

    def run_node_ctx_block(self, jac_ast):
        """
        node_ctx_block: name_list code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in self.run_name_list(kid[0]):
            if self.get_arch_for(self.current_node).is_instance(i):
                self.run_code_block(kid[1])
                return

    def run_walk_entry_block(self, jac_ast):
        """
        walk_entry_block: KW_WITH KW_ENTRY code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        if self.current_step == 0:
            self.in_entry_exit = True
            self.run_code_block(kid[2])
            self.in_entry_exit = False

    def run_walk_exit_block(self, jac_ast):
        """
        walk_exit_block: KW_WITH KW_EXIT code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        self._stopped = None
        if len(self.next_node_ids) == 0:
            self.in_entry_exit = True
            self.run_code_block(kid[2])
            self.in_entry_exit = False

    def run_walk_activity_block(self, jac_ast):
        """
        walk_activity_block: KW_WITH KW_ACTIVITY code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        self.run_code_block(kid[2])

    def run_walker_action(self, jac_ast):
        """
        walker_action:
            ignore_action
            | take_action
            | KW_DISENGAGE (report_action | SEMI);
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "KW_DISENGAGE":
            if kid[1].name == "report_action":
                self.run_report_action(kid[1])
            self._stopped = "stop"
            self.next_node_ids.remove_all()
        else:
            expr_func = getattr(self, f"run_{kid[0].name}")
            expr_func(kid[0])

    def run_ignore_action(self, jac_ast):
        """
        ignore_action: KW_IGNORE expression SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_expression(kid[1]).value
        if isinstance(result, node):
            self.ignore_node_ids.add_obj(result)
        elif isinstance(result, jac_set):
            self.ignore_node_ids.add_obj_list(result)
        else:
            self.rt_error(f"{result} is not ignorable type (i.e., nodes)", kid[1])

    def run_take_action(self, jac_ast):
        """
        take_action:
            KW_TAKE expression (SEMI | else_stmt);
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_expression(kid[1]).value
        before = len(self.next_node_ids)
        if isinstance(result, node):
            self.next_node_ids.add_obj(result, allow_dups=True)
        elif isinstance(result, jac_set):
            self.next_node_ids.add_obj_list(result, allow_dups=True)
        elif result:
            self.rt_error(f"{result} is not destination type (i.e., nodes)", kid[1])
        after = len(self.next_node_ids)
        if before >= after and kid[2].name == "else_stmt":
            self.run_else_stmt(kid[2])
        after = len(self.next_node_ids)

    def run_preset_in_out(self, jac_ast, obj, act):
        """
        preset_in_out:
            DBL_COLON expr_list? (DBL_COLON | COLON_OUT expression);

        obj: The node or edge with preset
        act: The action associated with preset
        """
        kid = self.set_cur_ast(jac_ast)
        param_list = []
        m = interp(parent_override=self.parent(), caller=self)
        arch = self.get_arch_for(obj)
        m.push_scope(
            jac_scope(parent=self, has_obj=obj, action_sets=[arch.activity_action_ids])
        )
        m._jac_scope.set_agent_refs(cur_node=self.current_node, cur_walker=self)

        if kid[1].name == "expr_list":
            param_list = m.run_expr_list(kid[1]).value
        try:
            result = act.trigger(param_list, self._jac_scope, self)
        except Exception as e:
            self.rt_error(f"Internal Exception: {e}", m._cur_jac_ast)
            result = None
        if kid[-1].name == "expression":
            dest = m.run_expression(kid[-1])
            dest.value = result
            dest.write(kid[-1])

    # Helper Functions ##################
    def auto_trigger_node_actions(self, nd, act_list):
        already_executed = []  # handles inhereted duplicates, (overriding)
        for i in act_list.obj_list():
            if (
                i.access_list
                and self.name not in i.access_list
                or i.name in already_executed
            ):
                continue
            if i.preset_in_out:
                self.run_preset_in_out(jac_ir_to_ast(i.preset_in_out), nd, i)
            else:
                self.call_ability(nd=nd, name=i.name, act_list=act_list)
            if not i.preset_in_out:  # All preset in and outs get executed
                already_executed.append(i.name)

    def viable_nodes(self):
        """Returns all nodes that shouldnt be ignored"""
        ret = jac_set()
        for i in self.current_node.attached_nodes():
            if i not in self.ignore_node_ids.obj_list():
                ret.add_obj(i)
        return ret

    def scope_and_run(self, jac_ast, run_func):
        """
        Helper to run ast elements with execution scope added
        (Useful for running arbitrary code blocks as one-offs)
        """
        arch = self.get_arch_for(self.current_node)
        self.push_scope(
            jac_scope(
                parent=self,
                has_obj=self,
                action_sets=[self.activity_action_ids, arch.activity_action_ids],
            )
        )
        self._jac_scope.set_agent_refs(cur_node=self.current_node, cur_walker=self)

        run_func(jac_ast)
        self.pop_scope()



"""
Sentinel api functions as a mixin
"""
from jaseci.api.interface import interface
from jaseci.utils.id_list import id_list
from jaseci.actor.sentinel import sentinel
from jaseci.utils.utils import b64decode_str
import uuid


class sentinel_api:
    """
    Sentinel APIs
    """

    def __init__(self):
        self.active_snt_id = None
        self.sentinel_ids = id_list(self)

    @interface.private_api(cli_args=["code"])
    def sentinel_register(
        self,
        name: str = "default",
        code: str = "",
        code_dir: str = "./",
        mode: str = "default",
        encoded: bool = False,
        auto_run: str = "init",
        auto_gen_graph: bool = True,
        ctx: dict = {},
        set_active: bool = True,
    ):
        """
        Create blank or code loaded sentinel and return object
        Auto_run is the walker to execute on register (assumes active graph
        is selected)
        """
        snt = self.sentinel_ids.get_obj_by_name(name, silent=True)
        new_gph = None
        if not snt:
            snt = sentinel(m_id=self._m_id, h=self._h, name=name)
            self.sentinel_ids.add_obj(snt)
            if auto_gen_graph:
                new_gph = self.graph_create(set_active=True)
        if code:
            self.sentinel_set(
                code=code, code_dir=code_dir, encoded=encoded, snt=snt, mode=mode
            )
            if not snt.is_active:
                return {
                    "response": "Error in jac code",
                    "errors": snt.errors + snt.runtime_errors,
                    "success": False,
                }
        if snt.walker_ids.has_obj_by_name(auto_run) and self.active_gph_id:
            nd = self._h.get_obj(self._m_id, uuid.UUID(self.active_gph_id))
            self.walker_run(name=auto_run, nd=nd, ctx=ctx, snt=snt)
        if set_active:
            self.sentinel_active_set(snt)
        self.extract_snt_aliases(snt)
        if new_gph:
            return [snt.serialize(), new_gph]
        return [snt.serialize()]

    @interface.private_api()
    def sentinel_pull(self, set_active: bool = True, on_demand: bool = True):
        """
        Copies global sentinel to local master
        """
        glob_id = self._h.get_glob("GLOB_SENTINEL")
        if not glob_id:
            return {"response": "No global sentinel is available!", "success": False}
        g_snt = self._h.get_obj(self._m_id, uuid.UUID(glob_id)).duplicate()

        snt = self.sentinel_ids.get_obj_by_name(g_snt.name, silent=True)
        if not snt:
            snt = sentinel(m_id=self._m_id, h=self._h, name=g_snt.name)
            self.sentinel_ids.add_obj(snt)
        elif on_demand and snt.is_active:
            return {"response": f"{snt} already active!", "success": True}
        if set_active:
            self.sentinel_active_set(snt)
        return self.sentinel_set(code=g_snt.code_ir, snt=snt, mode="ir")

    @interface.private_api()
    def sentinel_get(
        self, snt: sentinel = None, mode: str = "default", detailed: bool = False
    ):
        """
        Get a sentinel rendered with specific mode
        Valid modes: {default, code, ir, }
        """
        if mode == "code":
            return snt._jac_ast.get_text()
        elif mode == "ir":
            return snt.ir_dict()
        else:
            return snt.serialize(detailed=detailed)

    @interface.private_api(cli_args=["code"])
    def sentinel_set(
        self,
        code: str,
        code_dir: str = "./",
        encoded: bool = False,
        snt: sentinel = None,
        mode: str = "default",
    ):
        """
        Set code/ir for a sentinel, only replaces walkers/archs in sentinel
        Valid modes: {code, ir, }
        """
        if encoded:
            code = b64decode_str(code)
        if mode not in ["code", "default", "ir"]:
            return {"response": f"Invalid mode to set {snt}", "success": False}
        snt.register_code(code, dir=code_dir, mode=mode)
        snt.propagate_access()

        if snt.is_active:
            self.extract_snt_aliases(snt)
            return {"response": f"{snt} registered and active!", "success": True}
        else:
            return {
                "response": f"{snt} code issues encountered!",
                "success": False,
                "errors": snt.errors + snt.runtime_errors,
            }

    @interface.private_api()
    def sentinel_list(self, detailed: bool = False):
        """
        Provide complete list of all sentinel objects
        """
        snts = []
        for i in self.sentinel_ids.obj_list():
            snts.append(i.serialize(detailed=detailed))
        return snts

    @interface.private_api()
    def sentinel_test(self, snt: sentinel = None, detailed: bool = False):
        """
        Run battery of test cases within sentinel and provide result
        """
        return snt.run_tests(detailed=detailed)

    @interface.private_api(cli_args=["snt"])
    def sentinel_active_set(self, snt: sentinel):
        """
        Sets the default sentinel master should use
        """
        self.active_snt_id = snt.jid
        self.alias_register("active:sentinel", snt.jid)
        return [f"Sentinel {snt.id} set as default"]

    @interface.private_api()
    def sentinel_active_unset(self):
        """
        Unsets the default sentinel master should use
        """
        self.active_snt_id = None
        self.alias_delete("active:sentinel")
        return ["Default sentinel unset"]

    @interface.private_api()
    def sentinel_active_global(self, detailed: bool = False):
        """
        Sets the default master sentinel to the global sentinel
        Exclusive OR with pull strategy
        """
        ret = {"success": False, "sentinel": None}
        glob_id = self._h.get_glob("GLOB_SENTINEL")
        if not glob_id:
            ret["response"] = "No global sentinel is available!"
        else:
            self.active_snt_id = "global"  # Resolved in interface
            self.alias_register("active:sentinel", glob_id)
            ret["sentinel"] = self._h.get_obj(self._m_id, uuid.UUID(glob_id)).serialize(
                detailed=detailed
            )
            ret["success"] = True
            ret["response"] = f"Global sentinel {glob_id} set as default"
        return ret

    @interface.private_api()
    def sentinel_active_get(self, detailed: bool = False):
        """
        Returns the default sentinel master is using
        """
        id = self.active_snt_id
        if id == "global":
            id = self._h.get_glob("GLOB_SENTINEL")
        if not id:
            return {"response": "No default sentinel is selected!"}
        else:
            default = self._h.get_obj(self._m_id, uuid.UUID(id))
            return default.serialize(detailed=detailed)

    @interface.private_api(cli_args=["snt"])
    def sentinel_delete(self, snt: sentinel):
        """
        Permanently delete sentinel with given id
        """
        self.remove_snt_aliases(snt)
        if self.active_snt_id == snt.jid:
            self.sentinel_active_unset()
        self.sentinel_ids.destroy_obj(snt)
        return [f"Sentinel {snt.id} successfully deleted"]

    def active_snt(self):
        return self._h.get_obj(self._m_id, uuid.UUID(self.active_snt_id))

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.sentinel_ids.obj_list():
            i.destroy()



"""
Code generator for jac code in AST form
"""
from jaseci.jac.machine.op_codes import op
from jaseci.jac.machine.builtin_refs import ref


class code_gen:
    """Shared code generator class across both sentinels and walkers"""

    def __init__(self):
        self.machine_code = []
        self.label = []

    def gen_attr_stmt(self, jac_ast, obj):
        """
        attr_stmt: has_stmt | can_stmt;
        """
        kid = jac_ast.kid
        if kid[0].name == "has_stmt":
            self.gen_has_stmt(kid[0], obj)
        elif kid[0].name == "can_stmt":
            self.gen_can_stmt(kid[0], obj)

    def gen_has_stmt(self, jac_ast, obj):
        """
        has_stmt:
                KW_HAS KW_PRIVATE? KW_ANCHOR? has_assign
                (COMMA has_assign)* SEMI;
        """
        kid = jac_ast.kid
        kid = kid[1:]
        is_private = False
        is_anchor = False
        while True:
            if kid[0].name == "KW_PRIVATE":
                kid = kid[1:]
                is_private = True
            if kid[0].name == "KW_ANCHOR":
                kid = kid[1:]
                is_anchor = True
            self.gen_has_assign(kid[0], obj, is_private, is_anchor)
            kid = kid[1:]
            if not len(kid) or kid[0].name != "COMMA":
                break
            else:
                kid = kid[1:]

    def gen_has_assign(self, jac_ast, obj, is_private, is_anchor):
        """
        has_assign: NAME | NAME EQ expression;
        """
        kid = jac_ast.kid
        var_name = kid[0].token_text()
        if len(kid) > 1:
            self.gen_expression(kid[2])
        if is_anchor:
            self.g_ins([op.SET_ANCHOR, obj, var_name])
        if var_name == "_private":
            self.gt_error("Has variable name of `_private` not allowed!", kid[0])
        else:
            self.g_ins([op.CREATE_CTX_VAR, obj, var_name, ref.RESULT_OUT, is_private])

    # def gen_can_stmt(self, jac_ast, obj):
    #     """
    #     can_stmt:
    #         KW_CAN dotted_name preset_in_out? event_clause? (
    #             COMMA dotted_name preset_in_out? event_clause?
    #         )* SEMI
    #         | KW_CAN NAME event_clause? code_block;
    #     """
    #     kid = jac_ast.kid
    #     kid = kid[1:]
    #     l_out = self.next_lab()
    #     self.g_ins([op.B_A, l_out])
    #     while True:
    #         l_act_code = self.next_lab()
    #         self.g_lab(l_act_code)
    #         action_type = 'activity'
    #         preset_in_out = {'input': [], 'output': None}
    #         if (kid[0].name == 'NAME'):
    #             action_name = kid[0].token_text()
    #         else:
    #             action_name = self.gen_dotted_name(kid[0])
    #         kid = kid[1:]
    #         if(len(kid) > 0 and kid[0].name == 'preset_in_out'):
    #             self.gen_preset_in_out(kid[0], obj)
    #             self.g_ins([op.SET_REF_VAR, ref.TMP1, ref.RESULT_OUT])
    #             kid = kid[1:]
    #         if(len(kid) > 0 and kid[0].name == 'event_clause'):
    #             action_type = self.gen_event_clause(kid[0])
    #             kid = kid[1:]
    #         if (not isinstance(obj, node)):  # only nodes have on entry/exit
    #             action_type = 'activity'
    #         if (kid[0].name == 'code_block'):

    #             getattr(obj, f"{action_type}_action_ids").add_obj(
    #                 action(
    #                     h=self._h,
    #                     name=action_name,
    #                     value=kid[0],
    #                     preset_in_out=preset_in_out,
    #                     is_lib=False
    #                 )
    #             )
    #             break
    #         else:
    #             func_link = \
    #                 self.get_builtin_action(action_name, jac_ast)
    #             if(func_link):
    #                 getattr(obj, f"{action_type}_action_ids").add_obj(
    #                     action(
    #                         h=self._h,
    #                         name=action_name,
    #                         value=func_link,
    #                         preset_in_out=preset_in_out
    #                     )
    #                 )
    #         if(not len(kid) or kid[0].name != 'COMMA'):
    #             break
    #         else:
    #             kid = kid[1:]
    #     self.g_lab(l_out)

    # def gen_event_clause(self, jac_ast):
    #     """
    #     event_clause: KW_WITH (KW_ENTRY | KW_EXIT | KW_ACTIVITY);
    #     """
    #     kid = jac_ast.kid
    #     return kid[1].token_text()

    # def run_preset_in_out(self, jac_ast, obj):
    #     """
    #     preset_in_out: DBL_COLON NAME (COMMA NAME)* (COLON_OUT NAME)?;
    #     """
    #     kid = jac_ast.kid
    #     result = {'input': [], 'output': None}
    #     for i in kid:
    #         if (i.name == 'NAME'):
    #             if (i.token_text() not in obj.context.keys()):
    #                 self.rt_error(f"No context for preset param {i}", i)
    #             else:
    #                 prm = jac_value(obj, i.token_text())
    #                 result['input'].append(prm)
    #     if (kid[-2].name == 'COLON_OUT'):
    #         result['input'].pop()
    #         result['output'] = jac_value(obj, kid[-1].token_text())
    #     return result

    # def run_code_block(self, jac_ast):
    #     """
    #     code_block: LBRACE statement* RBRACE | COLON statement;
    #     TODO: Handle breaks and continues
    #     """
    #     kid = jac_ast.kid
    #     for i in kid:
    #         if (self._loop_ctrl):
    #             if (self._loop_ctrl == 'continue'):
    #                 self._loop_ctrl = None
    #             return
    #         if(i.name == 'statement'):
    #             self.run_statement(jac_ast=i)

    # def run_node_ctx_block(self, jac_ast):
    #     """
    #     node_ctx_block: NAME (COMMA NAME)* code_block;
    #     """
    #     kid = jac_ast.kid
    #     while(kid[0].name != 'code_block'):
    #         if (self.current_node.name == kid[0].token_text()):
    #             self.run_code_block(kid[-1])
    #             return
    #         kid = kid[1:]

    # def run_statement(self, jac_ast):
    #     """
    #     statement:
    #         code_block
    #         | node_ctx_block
    #         | expression SEMI
    #         | if_stmt
    #         | for_stmt
    #         | while_stmt
    #         | ctrl_stmt SEMI
    #         | report_action
    #         | walker_action;
    #     """
    #     if (self._stopped):
    #         return
    #     kid = jac_ast.kid
    #     if(not hasattr(self, f'run_{kid[0].name}')):
    #         self.rt_error(
    #             f'This scope cannot execute the statement '
    #             f'"{kid[0].get_text()}" of type {kid[0].name}',
    #             kid[0])
    #         return
    #     stmt_func = getattr(self, f'run_{kid[0].name}')
    #     stmt_func(kid[0])

    # def run_if_stmt(self, jac_ast):
    #     """
    #     if_stmt: KW_IF expression code_block (elif_stmt)* (else_stmt)?;
    #     """
    #     kid = jac_ast.kid
    #     if(self.run_expression(kid[1])):
    #         self.run_code_block(kid[2])
    #         return
    #     kid = kid[3:]
    #     if(len(kid)):
    #         while True:
    #             if(kid[0].name == 'elif_stmt'):
    #                 if(self.run_elif_stmt(kid[0])):
    #                     return
    #             elif(kid[0].name == 'else_stmt'):
    #                 self.run_else_stmt(kid[0])
    #                 return
    #             kid = kid[1:]
    #             if(not len(kid)):
    #                 break

    # def run_elif_stmt(self, jac_ast):
    #     """
    #     elif_stmt: KW_ELIF expression code_block;
    #     """
    #     kid = jac_ast.kid
    #     if(self.run_expression(kid[1])):
    #         self.run_code_block(kid[2])
    #         return True
    #     else:
    #         return False

    # def run_else_stmt(self, jac_ast):
    #     """
    #     else_stmt: KW_ELSE code_block;
    #     """
    #     kid = jac_ast.kid
    #     self.run_code_block(kid[1])

    # def run_for_stmt(self, jac_ast):
    #     """
    #     for_stmt:
    #         KW_FOR expression KW_TO expression KW_BY expression code_block
    #         | KW_FOR NAME KW_IN expression code_block;
    #     """
    #     kid = jac_ast.kid
    #     loops = 0
    #     if(kid[1].name == 'expression'):
    #         self.run_expression(kid[1])
    #         while self.run_expression(kid[3]):
    #             self.run_code_block(kid[6])
    #             loops += 1
    #             if (self._loop_ctrl == 'break'):
    #                 self._loop_ctrl = None
    #                 break
    #             self.run_expression(kid[5])
    #             if(loops > self._loop_limit):
    #                 self.rt_error(f'Hit loop limit, breaking...', kid[0])
    #     else:
    #         var_name = kid[1].token_text()
    #         lst = self.run_expression(kid[3])
    #         # should check that lst is list here
    #         for i in lst:
    #             self._jac_scope.set_live_var(var_name, i, [], kid[3])
    #             self.run_code_block(kid[4])
    #             loops += 1
    #             if (self._loop_ctrl == 'break'):
    #                 self._loop_ctrl = None
    #                 break
    #             if(loops > self._loop_limit):
    #                 self.rt_error(f'Hit loop limit, breaking...', kid[0])

    # def run_while_stmt(self, jac_ast):
    #     """
    #     while_stmt: KW_WHILE expression code_block;
    #     """
    #     kid = jac_ast.kid
    #     loops = 0
    #     while self.run_expression(kid[1]):
    #         self.run_code_block(kid[2])
    #         loops += 1
    #         if (self._loop_ctrl == 'break'):
    #             self._loop_ctrl = None
    #             break
    #         if(loops > self._loop_limit):
    #             self.rt_error(f'Hit loop limit, breaking...', kid[0])

    # def run_ctrl_stmt(self, jac_ast):
    #     """
    #     ctrl_stmt: KW_CONTINUE | KW_BREAK | KW_SKIP;
    #     """
    #     kid = jac_ast.kid
    #     if (kid[0].name == 'KW_SKIP'):
    #         self._stopped = 'skip'
    #     elif (kid[0].name == 'KW_BREAK'):
    #         self._loop_ctrl = 'break'
    #     elif (kid[0].name == 'KW_CONTINUE'):
    #         self._loop_ctrl = 'continue'

    # def run_report_action(self, jac_ast):
    #     """
    #     report_action: KW_REPORT expression SEMI;
    #     """
    #     kid = jac_ast.kid
    #     report = self.run_expression(kid[1])
    #     report = self._jac_scope.report_deep_serialize(report)
    #     if(not is_jsonable(report)):
    #         self.rt_error(f'Report not Json serializable', kid[0])
    #     self.report.append(report)

    # def run_expression(self, jac_ast):
    #     """
    #     expression: assignment | connect;
    #     """
    #     kid = jac_ast.kid
    #     expr_func = getattr(self, f'run_{kid[0].name}')
    #     return expr_func(kid[0])

    # def run_assignment(self, jac_ast, assign_scope=None):
    #     """
    #     assignment:
    #         dotted_name index* EQ expression
    #         | inc_assign
    #         | copy_assign;

    #     """
    #     kid = jac_ast.kid
    #     if (len(kid) < 2):
    #         if (assign_scope is not None):
    #             self.rt_error("Can only use '=' with spawn", kid[0])
    #         assign_func = getattr(self, f'run_{kid[0].name}')
    #         return assign_func(kid[0])
    #     var_name = self.run_dotted_name(kid[0])
    #     arr_idx = []
    #     for i in kid:
    #         if(i.name == 'index'):
    #             arr_idx.append(self.run_index(i))
    #     result = self.run_expression(kid[-1])
    #     if (assign_scope is None):
    #         self._jac_scope.set_live_var(var_name, result, arr_idx, kid[0])
    #     else:
    #         if(isinstance(result, element)):
    #             result = result.id.urn
    #         assign_scope[var_name] = result
    #     return result

    # def run_inc_assign(self, jac_ast):
    #     """
    #     inc_assign:
    #             dotted_name index* (PEQ | MEQ | TEQ | DEQ) expression;
    #     """
    #     kid = jac_ast.kid
    #     var_name = self.run_dotted_name(kid[0])
    #     arr_idx = []
    #     for i in kid:
    #         if(i.name == 'index'):
    #             arr_idx.append(self.run_index(i))
    #     result = self._jac_scope.get_live_var(var_name, kid[0])
    #     if(kid[1].name == 'PEQ'):
    #         result = result + self.run_expression(kid[2])
    #     elif(kid[1].name == 'MEQ'):
    #         result = result - self.run_expression(kid[2])
    #     elif(kid[1].name == 'TEQ'):
    #         result = result * self.run_expression(kid[2])
    #     elif(kid[1].name == 'DEQ'):
    #         result = result / self.run_expression(kid[2])
    #     self._jac_scope.set_live_var(var_name, result, arr_idx, kid[0])
    #     return result

    # def run_copy_assign(self, jac_ast):
    #     """
    #     copy_assign: dotted_name index* CPY_EQ expression;
    #     """
    #     kid = jac_ast.kid
    #     var_name = self.run_dotted_name(kid[0])
    #     dest = self._jac_scope.get_live_var(var_name, kid[0])
    #     for i in kid:
    #         if(i.name == 'index'):
    #             dest = dest[self.run_index(i)]
    #     src = self.run_expression(kid[2])
    #     if (not self.rt_check_type(dest, node, kid[0]) or not
    #             self.rt_check_type(dest, node, kid[0])):
    #         self.rt_error("':=' only applies to nodes", kid[0])
    #         return dest
    #     if (dest.name != src.name):
    #         self.rt_error(f"Node arch {dest} don't match {src}!", kid[0])
    #         return dest
    #     for i in src.context.keys():
    #         if(i in dest.context.keys()):
    #             dest.context[i] = src.context[i]
    #     return dest

    # def run_connect(self, jac_ast):
    #     """
    #     connect: logical ( (NOT)? edge_ref expression)?;
    #     """
    #     kid = jac_ast.kid
    #     if (len(kid) < 2):
    #         return self.run_logical(kid[0])
    #     base = self.run_logical(kid[0])
    #     target = self.run_expression(kid[-1])
    #     self.rt_check_type(base, node, kid[0])
    #     self.rt_check_type(target, [node, jac_set], kid[-1])
    #     if (kid[1].name == 'NOT'):
    #         # TODO: IF JACSET IS TARGET APLLY TO ALL MEMEBERS OF JACSET
    #         # Line below PARTIALLY generalizes disconnect NEEDS REVIEW
    #         if(isinstance(target, node)):
    #             base.detach_edges(target,
    #                               self.run_edge_ref(kid[2]).obj_list())
    #         elif(isinstance(target, jac_set)):
    #             for i in target.obj_list():
    #                 base.detach_edges(i,
    #                                   self.run_edge_ref(kid[2]).obj_list())
    #         else:
    #             logger.critical(
    #                 f'Cannot connect to {target} of type {type(target)}!')
    #         target = base
    #     else:
    #         direction = kid[1].kid[0].name
    #         if(isinstance(target, node)):
    #             target = jac_set(self, in_list=[target.jid])
    #         if(isinstance(target, jac_set)):
    #             for i in target.obj_list():
    #                 use_edge = self.run_edge_ref(kid[1], is_spawn=True)
    #                 if (direction == 'edge_from'):
    #                     base.attach_inbound(i, [use_edge])
    #                 elif (direction == 'edge_to'):
    #                     base.attach_outbound(i, [use_edge])
    #                 else:
    #                     base.attach_bidirected(i, [use_edge])
    #     return target

    # def run_logical(self, jac_ast):
    #     """
    #     logical: compare ((KW_AND | KW_OR) compare)*;
    #     """
    #     kid = jac_ast.kid
    #     result = self.run_compare(kid[0])
    #     kid = kid[1:]
    #     while (kid):
    #         if (kid[0].name == 'KW_AND'):
    #             if (result):
    #                 result = result and self.run_compare(kid[1])
    #         elif (kid[0].name == 'KW_OR'):
    #             if (not result):
    #                 result = result or self.run_compare(kid[1])
    #         kid = kid[2:]
    #         if(not kid):
    #             break
    #     return result

    # def run_compare(self, jac_ast):
    #     """
    #     compare:
    #         NOT compare
    #         | arithmetic (
    #             (EE | LT | GT | LTE | GTE | NE | KW_IN | nin) arithmetic
    #         )*;
    #     """
    #     kid = jac_ast.kid
    #     if(kid[0].name == 'NOT'):
    #         return not self.run_compare(kid[1])
    #     else:
    #         result = self.run_arithmetic(kid[0])
    #         kid = kid[1:]
    #         while (kid):
    #             other_res = self.run_arithmetic(kid[1])
    #             if(kid[0].name == 'EE'):
    #                 result = result == other_res
    #             elif(kid[0].name == 'LT'):
    #                 result = result < other_res
    #             elif(kid[0].name == 'GT'):
    #                 result = result > other_res
    #             elif(kid[0].name == 'LTE'):
    #                 result = result <= other_res
    #             elif(kid[0].name == 'GTE'):
    #                 result = result >= other_res
    #             elif(kid[0].name == 'NE'):
    #                 result = result != other_res
    #             elif(kid[0].name == 'KW_IN'):
    #                 result = result in other_res
    #             elif(kid[0].name == 'nin'):
    #                 result = result not in other_res
    #             kid = kid[2:]
    #             if(not kid):
    #                 break
    #         return result

    # def run_arithmetic(self, jac_ast):
    #     """
    #     arithmetic: term ((PLUS | MINUS) term)*;
    #     """
    #     kid = jac_ast.kid
    #     result = self.run_term(kid[0])
    #     kid = kid[1:]
    #     while (kid):
    #         other_res = self.run_term(kid[1])
    #         if(kid[0].name == 'PLUS'):
    #             result = result + other_res
    #         elif(kid[0].name == 'MINUS'):
    #             result = result - other_res
    #         kid = kid[2:]
    #         if(not kid):
    #             break
    #     return result

    # def run_term(self, jac_ast):
    #     """
    #     term: factor ((MUL | DIV | MOD) factor)*;
    #     """
    #     kid = jac_ast.kid
    #     result = self.run_factor(kid[0])
    #     kid = kid[1:]
    #     while (kid):
    #         other_res = self.run_factor(kid[1])
    #         if(kid[0].name == 'MUL'):
    #             result = result * other_res
    #         elif(kid[0].name == 'DIV'):
    #             result = result / other_res
    #         elif(kid[0].name == 'MOD'):
    #             result = result % other_res
    #         kid = kid[2:]
    #         if(not kid):
    #             break
    #     return result

    # def run_factor(self, jac_ast):
    #     """
    #     factor: (PLUS | MINUS) factor | power;
    #     """
    #     kid = jac_ast.kid
    #     if(kid[0].name == 'power'):
    #         return self.run_power(kid[0])
    #     else:
    #         result = self.run_factor(kid[1])
    #         if(kid[0].name == 'MINUS'):
    #             result = -(result)
    #         return result

    # def run_power(self, jac_ast):
    #     """
    #     power: func_call (POW factor)* | func_call index+;
    #     """
    #     kid = jac_ast.kid
    #     result = self.run_func_call(kid[0])
    #     kid = kid[1:]
    #     if(len(kid) < 1):
    #         return result
    #     elif(kid[0].name == 'POW'):
    #         while (kid):
    #             result = result ** self.run_factor(kid[1])
    #             kid = kid[2:]
    #             if(not kid):
    #                 break
    #         return result
    #     elif (kid[0].name == "index"):
    #         if(isinstance(result, list) or isinstance(result, dict)):
    #             for i in kid:
    #                 if(i.name == 'index'):
    #                     result = result[self.run_index(i)]
    #             result = self._jac_scope.reference_to_value(result)
    #             return result
    #         else:
    #             self.rt_error(f'Cannot index into {result}'
    #                           f' of type {type(result)}!',
    #                           kid[0])
    #             return 0

    # def run_func_call(self, jac_ast):
    #     """
    #     func_call:
    #         atom (LPAREN (expression (COMMA expression)*)? RPAREN)?
    #         | atom DOT func_built_in
    #         | atom? DBL_COLON NAME;
    #     """
    #     kid = jac_ast.kid
    #     atom_res = self._jac_scope.has_obj
    #     if (kid[0].name == 'atom'):
    #         atom_res = self.run_atom(kid[0])
    #         kid = kid[1:]
    #     if(len(kid) < 1):
    #         return atom_res
    #     elif(kid[0].name == 'DOT'):
    #         return self.run_func_built_in(atom_res, kid[1])
    #     elif (kid[0].name == 'DBL_COLON'):
    #         m = interp(owner_override=self.owner())
    #         m.push_scope(jac_scope(owner=self,
    #                                has_obj=atom_res,
    #                                action_sets=[atom_res.activity_action_ids]))
    #         m.run_code_block(atom_res.activity_action_ids.get_obj_by_name(
    #             kid[1].token_text()).value)
    #         self.report = self.report + m.report
    #         return atom_res
    #     elif(kid[0].name == "LPAREN"):
    #         param_list = []
    #         kid = kid[1:]
    #         while True:
    #             if(kid[0].name == 'RPAREN'):
    #                 break
    #             param_list.append(self.run_expression(kid[0]))
    #             kid = kid[1:]
    #             if (kid[0].name == 'COMMA'):
    #                 kid = kid[1:]
    #         if (isinstance(atom_res, action)):
    #             return atom_res.trigger(param_list)
    #         else:
    #             self.rt_error(f'Unable to execute function {atom_res}',
    #                           kid[0])

    # def run_func_built_in(self, atom_res, jac_ast):
    #     """
    #     func_built_in:
    #         | KW_LENGTH
    #         | KW_KEYS
    #         | KW_EDGE
    #         | KW_NODE
    #         | KW_DESTROY LPAREN expression RPAREN;
    #     """
    #     kid = jac_ast.kid
    #     if (kid[0].name == "KW_LENGTH"):
    #         if(isinstance(atom_res, list)):
    #             return len(atom_res)
    #         else:
    #             self.rt_error(f'Cannot get length of {atom_res}. Not List!',
    #                           kid[0])
    #             return 0
    #     elif (kid[0].name == "KW_KEYS"):
    #         if(isinstance(atom_res, dict)):
    #             return atom_res.keys()
    #         else:
    #             self.rt_error(f'Cannot get keys of {atom_res}. '
    #                           f'Not Dictionary!', kid[0])
    #             return []
    #     elif (kid[0].name == "KW_EDGE"):
    #         if(isinstance(atom_res, node)):
    #             return self.obj_set_to_jac_set(
    #                 self.current_node.attached_edges(atom_res))
    #         elif(isinstance(atom_res, edge)):
    #             return atom_res
    #         elif(isinstance(atom_res, jac_set)):
    #             res = jac_set(self)
    #             for i in atom_res.obj_list():
    #                 if(isinstance(i, edge)):
    #                     res.add_obj(i)
    #                 elif(isinstance(i, node)):
    #                     res += self.obj_set_to_jac_set(
    #                         self.current_node.attached_edges(i))
    #             return res
    #         else:
    #             self.rt_error(f'Cannot get edges from {atom_res}. '
    #                           f'Type {type(atom_res)} invalid', kid[0])
    #             return atom_res
    #     # may want to remove 'here" node from return below
    #     elif (kid[0].name == "KW_NODE"):
    #         if(isinstance(atom_res, node)):
    #             return atom_res
    #         elif(isinstance(atom_res, edge)):
    #             return self.obj_set_to_jac_set(atom_res.nodes())
    #         elif(isinstance(atom_res, jac_set)):
    #             res = jac_set(self)
    #             for i in atom_res.obj_list():
    #                 if(isinstance(i, edge)):
    #                     res.add_obj(i.to_node())
    #                     res.add_obj(i.from_node())
    #                 elif(isinstance(i, node)):
    #                     res.add_obj(i)
    #             return res
    #         else:
    #             self.rt_error(f'Cannot get edges from {atom_res}. '
    #                           f'Type {type(atom_res)} invalid', kid[0])
    #             return atom_res
    #     elif (kid[0].name == "KW_DESTROY"):
    #         idx = self.run_expression(kid[2])
    #         if (isinstance(atom_res, list) and isinstance(idx, int)):
    #             del atom_res[idx]
    #             return atom_res
    #         else:
    #             self.rt_error(f'Cannot remove index {idx} from {atom_res}.',
    #                           kid[0])
    #             return atom_res

    # def run_atom(self, jac_ast):
    #     """
    #     atom:
    #         INT
    #         | FLOAT
    #         | STRING
    #         | BOOL
    #         | array_ref
    #         | node_edge_ref
    #         | list_val
    #         | dotted_name
    #         | LPAREN expression RPAREN
    #         | spawn
    #         | DEREF expression;
    #     """
    #     kid = jac_ast.kid
    #     if(kid[0].name == 'INT'):
    #         return int(kid[0].token_text())
    #     elif(kid[0].name == 'FLOAT'):
    #         return float(kid[0].token_text())
    #     elif(kid[0].name == 'STRING'):
    #         return self.parse_str_token(kid[0].token_text())
    #     elif(kid[0].name == 'BOOL'):
    #         return bool(kid[0].token_text() == 'true')

    #     elif(kid[0].name == 'dotted_name'):
    #         return self._jac_scope.get_live_var(self.run_dotted_name(kid[0]),
    #                                             kid[0])
    #     elif(kid[0].name == 'LPAREN'):
    #         return self.run_expression(kid[1])
    #     elif (kid[0].name == 'DEREF'):
    #         result = self.run_expression(kid[1])
    #         if (self.rt_check_type(result, element, kid[1])):
    #             result = result.jid
    #         return result
    #     else:
    #         return getattr(self, f'run_{kid[0].name}')(kid[0])

    # def run_node_edge_ref(self, jac_ast):
    #     """
    #     node_edge_ref: node_ref | edge_ref (node_ref)?;
    #     """
    #     kid = jac_ast.kid
    #     is_nodeset = True
    #     if(kid[0].name == 'KW_NODE'):
    #         kid = kid[2:]
    #     if(kid[0].name == 'KW_EDGE'):
    #         kid = kid[2:]
    #         is_nodeset = False

    #     if(kid[0].name == 'node_ref'):
    #         if(is_nodeset):
    #             return self.run_node_ref(kid[0])
    #         else:
    #             return self.obj_set_to_jac_set(
    #                 self.current_node.attached_edges(
    #                     self.run_node_ref(kid[0])))
    #     elif (kid[0].name == 'edge_ref'):
    #         if(is_nodeset):
    #             result = self.edge_to_node_jac_set(self.run_edge_ref(kid[0]))
    #             if(len(kid) > 1 and kid[1].name == 'node_ref'):
    #                 result = result * self.run_node_ref(kid[1])
    #             return result
    #         else:
    #             result = self.run_edge_ref(kid[0])
    #             if(kid[1].name == 'node_ref'):
    #                 result = jac_set(
    #                     self, inlist=[i for i in result if i in
    #                                   self.obj_set_to_jac_set(
    #                                       self.current_node.attached_edges(
    #                                           self.run_node_ref(kid[1])))])
    #             return result

    # def run_node_ref(self, jac_ast, is_spawn=False):
    #     """
    #     node_ref: KW_NODE DBL_COLON NAME;
    #     """
    #     kid = jac_ast.kid
    #     if(not is_spawn):
    #         result = jac_set(self.owner())
    #         if (len(kid) > 1):
    #             for i in self.viable_nodes().obj_list():
    #                 if (i.name == kid[2].token_text()):
    #                     result.add_obj(i)
    #         else:
    #             result += self.viable_nodes()
    #     else:
    #         if(len(kid) > 1):
    #             result = self.owner().arch_ids.get_obj_by_name(
    #                 kid[2].token_text(), kind='node').run()
    #         else:
    #             result = node(h=self._h)
    #     return result

    # def run_walker_ref(self, jac_ast):
    #     """
    #     walker_ref: KW_WALKER DBL_COLON NAME;
    #     """
    #     kid = jac_ast.kid
    #     return self.owner().spawn_walker(kid[2].token_text())

    # def run_graph_ref(self, jac_ast):
    #     """
    #     graph_ref: KW_GRAPH DBL_COLON NAME;
    #     """
    #     kid = jac_ast.kid
    #     gph = self.owner().arch_ids.get_obj_by_name(
    #         kid[2].token_text(), kind='graph').run()
    #     return gph

    # def run_edge_ref(self, jac_ast, is_spawn=False):
    #     """
    #     edge_ref: edge_to | edge_from | edge_any;
    #     """
    #     kid = jac_ast.kid
    #     if(not is_spawn):
    #         expr_func = getattr(self, f'run_{kid[0].name}')
    #         return expr_func(kid[0])
    #     else:
    #         if(len(kid[0].kid) > 2):
    #             result = self.owner().arch_ids.get_obj_by_name(
    #                 kid[0].kid[2].token_text(), kind='edge').run()
    #         else:
    #             result = edge(h=self._h, kind='edge', name='generic')
    #         return result

    # def run_edge_to(self, jac_ast):
    #     """
    #     edge_to: '-' ('[' NAME ']')? '->';
    #     """
    #     kid = jac_ast.kid
    #     result = jac_set(self.owner())
    #     for i in self.current_node.outbound_edges() + \
    #             self.current_node.bidirected_edges():
    #         if (len(kid) > 2 and i.name != kid[2].token_text()):
    #             continue
    #         result.add_obj(i)
    #     return result

    # def run_edge_from(self, jac_ast):
    #     """
    #     edge_from: '<-' ('[' NAME ']')? '-';
    #     """
    #     kid = jac_ast.kid
    #     result = jac_set(self.owner())
    #     for i in self.current_node.inbound_edges() + \
    #             self.current_node.bidirected_edges():
    #         if (len(kid) > 2 and i.name != kid[2].token_text()):
    #             continue
    #         result.add_obj(i)
    #     return result

    # def run_edge_any(self, jac_ast):
    #     """
    #     edge_any: '<-' ('[' NAME ']')? '->';
    #     NOTE: these do not use strict bidirected semantic but any edge
    #     """
    #     kid = jac_ast.kid
    #     result = jac_set(self.owner())
    #     for i in self.current_node.attached_edges():
    #         if (len(kid) > 2 and i.name != kid[2].token_text()):
    #             continue
    #         result.add_obj(i)
    #     return result

    # def run_list_val(self, jac_ast):
    #     """
    #     list_val: LSQUARE (expression (COMMA expression)*)? RSQUARE;
    #     """
    #     kid = jac_ast.kid
    #     list_res = []
    #     for i in kid:
    #         if(i.name == 'expression'):
    #             list_res.append(self.run_expression(i))
    #     return list_res

    # def run_index(self, jac_ast):
    #     """
    #     index: LSQUARE expression RSQUARE;
    #     """
    #     kid = jac_ast.kid
    #     idx = self.run_expression(kid[1])
    #     if(not isinstance(idx, int) and not isinstance(idx, str)):
    #         self.rt_error(f'Index of type {type(idx)} not valid. '
    #                       f'Indicies must be an integer or string!', kid[1])
    #         return None
    #     return idx

    # def run_dict_val(self, jac_ast):
    #     """
    #     dict_val: LBRACE (kv_pair (COMMA kv_pair)*)? RBRACE;
    #     """
    #     kid = jac_ast.kid
    #     dict_res = {}
    #     for i in kid:
    #         if(i.name == 'kv_pair'):
    #             self.run_kv_pair(i, dict_res)
    #     return dict_res

    # def run_kv_pair(self, jac_ast, obj):
    #     """
    #     kv_pair: STRING COLON expression;
    #     """
    #     kid = jac_ast.kid
    #     obj[self.parse_str_token(kid[0].token_text())
    #         ] = self.run_expression(kid[2])

    # def run_spawn(self, jac_ast):
    #     """
    #     spawn: KW_SPAWN expression spawn_object;

    #     NOTE: spawn statements support locations that are either nodes or
    #     jac_sets
    #     """
    #     kid = jac_ast.kid
    #     if(kid[1].name == 'expression'):
    #         location = self.run_expression(kid[1])
    #         if(isinstance(location, node)):
    #             return self.run_spawn_object(kid[2], location)
    #         elif(isinstance(location, jac_set)):
    #             res = []
    #             for i in location.obj_list():
    #                 res.append(self.run_spawn_object(kid[2], i))
    #             return res
    #         else:
    #             self.rt_error(
    #                 f'Spawn can not occur on {type(location)}!', kid[1])
    #     else:
    #         return self.run_spawn_object(kid[1], None)

    # def run_spawn_object(self, jac_ast, location):
    #     """
    #     spawn_object: node_spawn | walker_spawn;
    #     """
    #     kid = jac_ast.kid
    #     expr_func = getattr(self, f'run_{kid[0].name}')
    #     return expr_func(kid[0], location)

    # def run_node_spawn(self, jac_ast, location):
    #     """
    #     node_spawn: edge_ref? node_ref spawn_ctx?;
    #     """
    #     kid = jac_ast.kid
    #     if(kid[0].name == 'node_ref'):
    #         ret_node = self.run_node_ref(kid[0], is_spawn=True)
    #     else:
    #         use_edge = self.run_edge_ref(kid[0], is_spawn=True)
    #         ret_node = self.run_node_ref(kid[1], is_spawn=True)
    #         direction = kid[0].kid[0].name
    #         if (direction == 'edge_from'):
    #             location.attach_inbound(ret_node, [use_edge])
    #         elif (direction == 'edge_to'):
    #             location.attach_outbound(ret_node, [use_edge])
    #         else:
    #             location.attach_bidirected(ret_node, [use_edge])
    #     if (kid[-1].name == 'spawn_ctx'):
    #         self.run_spawn_ctx(kid[-1], ret_node)
    #     return ret_node

    # def run_walker_spawn(self, jac_ast, location):
    #     """
    #     walker_spawn: walker_ref spawn_ctx?;
    #     """
    #     kid = jac_ast.kid
    #     walk = self.run_walker_ref(kid[0])
    #     walk.prime(location)
    #     if(len(kid) > 1):
    #         self.run_spawn_ctx(kid[1], walk)
    #     walk.run()
    #     ret = self._jac_scope.reference_to_value(walk.anchor_value())
    #     self.report = self.report + walk.report
    #     walk.destroy()
    #     return ret

    # def run_graph_spawn(self, jac_ast, location):
    #     """
    #     graph_spawn: edge_ref graph_ref;
    #     """
    #     kid = jac_ast.kid
    #     use_edge = self.run_edge_ref(kid[0], is_spawn=True)
    #     result = self.run_graph_ref(kid[1])
    #     direction = kid[0].kid[0].name
    #     if (direction == 'edge_from'):
    #         location.attach_inbound(result, [use_edge])
    #     elif (direction == 'edge_to'):
    #         location.attach_outbound(result, [use_edge])
    #     else:
    #         location.attach_bidirected(result, [use_edge])
    #     return result

    # def run_spawn_ctx(self, jac_ast, obj):
    #     """
    #     spawn_ctx: LPAREN (assignment (COMMA assignment)*)? RPAREN;
    #     """
    #     kid = jac_ast.kid
    #     for i in kid:
    #         if (i.name == 'assignment'):
    #             self.run_assignment(i, assign_scope=obj.context)

    # def gen_dotted_name(self, jac_ast):
    #     """
    #     dotted_name: NAME (DOT NAME)*;
    #     """
    #     kid = jac_ast.kid
    #     ret = ''
    #     while True:
    #         ret += kid[0].token_text()
    #         kid = kid[1:]
    #         if(not len(kid) or kid[0].name != 'DOT'):
    #             break
    #         if(kid[0].name == 'DOT'):
    #             ret += '.'
    #             kid = kid[1:]
    #     return ret

    # Helper Functions ##################
    def g_ins(self, ins):
        """Generates op with specified parameters"""
        self.machine_code.append(ins)

    def g_lab(self, name):
        """Inserts label in code"""
        self.machine_code.append(name)

    def next_lab(self):
        """Generate and return new label"""
        lab = f"L{len(self.label)}"
        self.label.append(lab)
        return lab



"""
General action base class with automation for hot loading
"""
from importlib.util import spec_from_file_location, module_from_spec
from jaseci.utils.utils import logger
from jaseci.actions.remote_actions import ACTIONS_SPEC_LOC
from jaseci.actions.remote_actions import serv_actions, mark_as_remote
import requests
import os
import sys
import inspect
import importlib

live_actions = {}


def jaseci_action(act_group=None, aliases=list(), allow_remote=False):
    """Decorator for Jaseci Action interface"""
    caller_globals = dict(inspect.getmembers(inspect.currentframe().f_back))[
        "f_globals"
    ]
    if allow_remote and "serv_actions" not in caller_globals:
        caller_globals["serv_actions"] = serv_actions

    def decorator_func(func):
        if allow_remote:
            mark_as_remote([func, act_group, aliases, caller_globals])
        return assimilate_action(func, act_group, aliases)

    return decorator_func


def assimilate_action(func, act_group=None, aliases=list()):
    """Helper for jaseci_action decorator"""
    act_group = [func.__module__.split(".")[-1]] if act_group is None else act_group
    live_actions[f"{'.'.join(act_group+[func.__name__])}"] = func
    for i in aliases:
        live_actions[f"{'.'.join(act_group+[i])}"] = func
    return func


def load_local_actions(file: str):
    """Load all jaseci actions from python file"""
    name = file.rstrip(".py")
    name = ".".join(name.split("/")[-2:])
    # Assumes parent folder of py file is a package for internal relative
    # imports, name is package.module and package path is added to sys path
    spec = spec_from_file_location(name, str(file))
    if spec is None:
        logger.error(f"Cannot hot load from action file {file}")
        return False
    else:
        module_dir = os.path.dirname(os.path.dirname(os.path.realpath(file)))
        if module_dir not in sys.path:
            sys.path.append(module_dir)
        spec.loader.exec_module(module_from_spec(spec))
        return True


def load_module_actions(mod):
    """Load all jaseci actions from python module"""
    if importlib.import_module(mod):
        return True
    return False


def load_preconfig_actions(hook):
    import json

    action_preload = hook.resolve_glob("ACTION_SETS", None)
    if action_preload:
        action_preload = json.loads(action_preload)
        for i in action_preload["local"]:
            load_local_actions(i)
        for i in action_preload["remote"]:
            load_remote_actions(i)
        for i in action_preload["module"]:
            load_module_actions(i)


def get_global_actions(hook):
    """
    Loads all global action hooks for use by Jac programs
    Attaches globals to mem_hook
    """
    from jaseci.attr.action import action
    import uuid

    global_action_list = []
    for i in live_actions.keys():
        if (
            i.startswith("std.")
            or i.startswith("file.")
            or i.startswith("net.")
            or i.startswith("rand.")
            or i.startswith("vector.")
            or i.startswith("request.")
            or i.startswith("date.")
            or i.startswith("jaseci.")
        ):
            global_action_list.append(
                action(
                    m_id=uuid.UUID(int=0).urn,
                    h=hook,
                    mode="public",
                    name=i,
                    value=i,
                    persist=False,
                ).jid
            )
    return global_action_list


def load_remote_actions(url):
    """Load all jaseci actions from live pod"""
    headers = {"content-type": "application/json"}
    try:
        spec = requests.get(url.rstrip("/") + ACTIONS_SPEC_LOC, headers=headers)
        spec = spec.json()
        for i in spec.keys():
            live_actions[i] = gen_remote_func_hook(url, i, spec[i])
        return True

    except Exception as e:
        logger.error(f"Cannot hot load remote actions at {url}: {e}")
        return False


def gen_remote_func_hook(url, act_name, param_names):
    """Generater for function calls for remote action calls"""

    def func(*args):
        params = {}
        for i in range(len(param_names)):
            if i < len(args):
                params[param_names[i]] = args[i]
            else:
                params[param_names[i]] = None
        act_url = f"{url.rstrip('/')}/{act_name.split('.')[-1]}"
        res = requests.post(
            act_url, headers={"content-type": "application/json"}, json=params
        )
        return res.json()

    return func



"""
Graph api functions as a mixin
"""
from jaseci.api.interface import interface
from jaseci.utils.id_list import id_list
from jaseci.graph.graph import graph
from jaseci.graph.node import node
from jaseci.actor.sentinel import sentinel
import uuid


class graph_api:
    """
    Graph APIs
    """

    def __init__(self):
        self.active_gph_id = None
        self.graph_ids = id_list(self)

    @interface.private_api()
    def graph_create(self, set_active: bool = True):
        """
        Create a graph instance and return root node graph object
        """
        gph = graph(m_id=self._m_id, h=self._h)
        self.graph_ids.add_obj(gph)
        if set_active:
            self.graph_active_set(gph)
        return gph.serialize()

    @interface.private_api()
    def graph_get(
        self, gph: graph = None, mode: str = "default", detailed: bool = False
    ):
        """
        Return the content of the graph with mode
        Valid modes: {default, dot, }
        """
        if mode == "dot":
            return gph.graph_dot_str()
        else:
            items = []
            for i in gph.get_all_nodes():
                items.append(i.serialize(detailed=detailed))
            for i in gph.get_all_edges():
                items.append(i.serialize(detailed=detailed))
            return items

    @interface.private_api()
    def graph_list(self, detailed: bool = False):
        """
        Provide complete list of all graph objects (list of root node objects)
        """
        gphs = []
        for i in self.graph_ids.obj_list():
            gphs.append(i.serialize(detailed=detailed))
        return gphs

    @interface.private_api(cli_args=["gph"])
    def graph_active_set(self, gph: graph):
        """
        Sets the default graph master should use
        """
        self.active_gph_id = gph.jid
        self.alias_register("active:graph", gph.jid)
        return [f"Graph {gph.id} set as default"]

    @interface.private_api()
    def graph_active_unset(self):
        """
        Unsets the default sentinel master should use
        """
        self.active_gph_id = None
        self.alias_delete("active:graph")
        return ["Default graph unset"]

    @interface.private_api()
    def graph_active_get(self, detailed: bool = False):
        """
        Returns the default graph master is using
        """
        if self.active_gph_id:
            default = self._h.get_obj(self._m_id, uuid.UUID(self.active_gph_id))
            return default.serialize(detailed=detailed)
        else:
            return {"success": False, "response": "No default graph is selected!"}

    @interface.private_api(cli_args=["gph"])
    def graph_delete(self, gph: graph):
        """
        Permanently delete graph with given id
        """
        if self.active_gph_id == gph.jid:
            self.graph_active_unset()
        self.graph_ids.destroy_obj(gph)
        return [f"Graph {gph.id} successfully deleted"]

    @interface.private_api(cli_args=["nd"])
    def graph_node_get(self, nd: node, ctx: list = None):
        """
        Returns value a given node
        """
        ret = {}
        nd_ctx = nd.serialize(detailed=True)["context"]
        if ctx:
            for i in nd_ctx.keys():
                if i in ctx:
                    ret[i] = nd_ctx[i]
        return ret

    @interface.private_api(cli_args=["nd"])
    def graph_node_set(self, nd: node, ctx: dict, snt: sentinel = None):
        """
        Assigns values to member variables of a given node using ctx object
        """
        temp_ref_nd = snt.run_architype(nd.name, kind="node", caller=self)
        nd.set_context(ctx=ctx, arch=temp_ref_nd)
        temp_ref_nd.destroy()
        return nd.serialize()

    @interface.cli_api(cli_args=["file"])
    def graph_walk(self, nd: node = None):
        cmd = ""
        while cmd not in ["quit", "q", "exit"]:
            print(
                "location - " + ":".join([nd.kind, nd.name, nd.jid.strip("urn:uuid:")])
            )
            cmd = input("graph_walk_mode > ")

    def active_gph(self):
        return self._h.get_obj(self._m_id, uuid.UUID(self.active_gph_id))

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.graph_ids.obj_list():
            i.destroy()



bug_check1 = """
    node state {
        has cand_intents = [];
        # Collect possible intents from a given conversation state
        can collect_intents {
            for i in -[transition]->.edge {
                here.cand_intents += [i.intent_label];
            }
        }
    }
    edge transition {has intent_label;}
    walker test_walker {
        state {
            here::collect_intents;
        }
    }
    graph test_graph {
        has anchor state_node;
        spawn {
            state_node = spawn node::state;
            state_2 = spawn node::state;
            state_node -[transition(intent_label =
                        "THIS IS AN INTENT_LABEL")]-> state_2;
        }
    }

    walker init {
        spawn here --> graph::test_graph;
        spawn --> walker::test_walker;
        report -->[0].cand_intents;
        report -->;
    }

    test "Test"
    with graph::test_graph by walker::test_walker {
        assert(here.cand_intents == ['THIS IS AN INTENT_LABEL']);
    }
    """


action_load_std_lib = """
    walker aload {
        report std.actload_local('jaseci/tests/infer.py');
    }
    """


globals = """
    global a = "testing";

    walker init {
        report global.a;
        global.a = 56;
        report global.a;
    }
    """

net_root_std_lib = """
    walker init {
        root {
            report [here.info['jid'], net.root().info['jid']];
            spawn here --> node::generic;
            take -->;
        }
        generic {
            report [here.info['jid'], net.root().info['jid']];
        }
    }
    """


or_stmt = """
    walker init {
        x = 3.4;
        y = "Hello";
        if(x == 3.45 or y == "Hello"):
            report [x, y];
    }
    """

nd_equals_error_correct_line = """
    node plain{ has name="joe", noname;
    can stuff with entry {std.out(name + noname); }}

    walker init {
        root {
            spawn here --> node::plain;
            take -->;
        }
        plain {
            if(here.details['name'] == 'joe'):
                report here.info;
        }
    }
    """

strange_ability_bug = """
    node plain {
        can show with entry {
            report "Showing";
        }
    }

    walker init {
        root {
            spawn here --> node::plain;
        }
    }

    walker travel {
        take -->;
    }
    """

node_inheritance = """
    node plain {
        has a=5, b=7, c=7, d=8;
        can x with entry {
            report "plain.x";
        }
        can y {
            report "plain.y";
        }
    }

    node plain2 {
        has c=70, d=80;
        can x with entry {
            report "plain2.x";
        }
        can y {
            report "plain2.y";
        }
    }

    node super:plain:plain2 {
        has a=55, c=7;
        can x with entry {
            ::plain:x;
            report here.context;
            report "super.x";
        }
        can y {
            ::plain2:y;
            report "super.y";
        }
    }

    walker init {
        root {
            a=spawn here --> node::super;
        }
        take -->;
        super {
            here::y;
        }
    }
    """

inherited_ref = """
    node plain {
        has a=5, b=7, c=7, d=8;
        can x with entry {
            report "plain.x";
        }
        can y {
            report "plain.y";
        }
        can z with entry {
            report "Z triggered";
        }
    }

    node plain2:plain {
        has c=70, d=80;
        can x with entry {
            report "plain2.x";
        }
        can y {
            report "plain2.y";
        }
        can z with entry {
            report "New Z triggered";
        }
    }

    node super:plain2 {
        has a=55, c=7;
        can x with entry {
            ::plain:x;
            report here.context;
            report "super.x";
            ::z;
        }
        can y {
            ::plain2:y;
            report "super.y";
        }
    }

    walker init {
        root {
            spawn here --> node::super;
            spawn here --> node::plain;
            spawn here --> node::plain2;
        }
        take --> node::plain;
        plain {
            report here;
        }
    }
    """

node_inheritance_chain_check = """
    node plain {
        has a=5, b=7, c=7, d=8;
        can x with entry {
            report "plain.x";
        }
        can y {
            report "plain.y";
        }
    }

    node plain2 {
        has c=70, d=80;
        can x with entry {
            report "plain2.x";
        }
        can y {
            report "plain2.y";
        }
    }

    node super:plain {
        has a=55, c=7;
        can x with entry {
            ::plain:x;
            report here.context;
            report "super.x";
        }
        can y {
            ::plain2:y;
            report "super.y";
        }
    }

    walker init {
        root {
            a=spawn here --> node::super;
        }
        take -->;
        super {
            here::y;
        }
    }
    """

global_reregistering = """
    node plain;

    global a = '556';

    walker init {
        root {
            spawn here - -> node::plain;
            spawn here - -> node::plain;
            spawn here - -> node::plain;
        }
        report global.a;
        take - ->;
    }
    """

vector_cos_sim_check = """
    node plain;

    walker init {
        a=[1,2,3];
        b=[4,5,6];
        report vector.cosine_sim(a,b);
    }
    """

multi_breaks = """
    node plain {
        has anchor val=0;
        can breakdance {
            for i=0 to i<10 by i+=1 {
                for j=0 to j<20 by j+=1:
                    if(j==12) {
                        val+=j;
                        break;
                    }
                report "here";
                if(i==6){
                    val+=i;
                    break;
                }
            }
            break;
            val+=100;
        }
    }

    walker init {
        nd=spawn here --> node::plain;
        nd::breakdance;
        nd::breakdance;
        report nd.val;
    }
    """

reffy_deref_check = """
    node plain{has expected_answer;}

    walker init {
        nd = spawn here --> node::plain;
        spawn here --> node::plain;

        report *&-->[0] == *&-->[1];

        actual_answer = -->[1];
        nd.expected_answer = &-->[1];
        report *&actual_answer == *nd.expected_answer;
    }
    """

vanishing_can_check = """
    node plain {
        has name;
        can infer.year_from_date;
    }

    walker init {
        root {
            take --> node::plain else {
                nd=spawn here --> node::plain;
                report nd.info['jid'];
                disengage;
            }
        }
        plain {
            report infer.year_from_date("2022-05-05");
        }
    }
    """

jasecilib_alias_list = """
    walker init {
        report jaseci.alias_list();
    }
    """

jasecilib_params = """
    walker init {
        report jaseci.graph_list(true);
    }
    """

jasecilib_create_user = """
    walker init {
        report jaseci.master_create("daman@gmail.com");
    }
    """



from jaseci.api.interface import interface
import stripe

stripe_test_key = "sk_test_4eC39HqLyjWDarjtT1zdp7dc"


class stripe_api:
    """
    Stripe APIs
    Set of APIs to expose jaseci stripe management
    """

    def __init__(self):
        stripe.api_key = self._h.resolve_glob("STRIPE_KEY", stripe_test_key)

    @interface.admin_api()
    def stripe_product_create(
        self, name: str = "VIP Plan", description: str = "Plan description"
    ):
        """create product"""
        try:
            return stripe.Product.create(namex=name, descriptionx=description)
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_product_price_set(
        self, productId: str, amount: float = 50, interval: str = "month"
    ):
        """modify product price"""
        try:
            return stripe.Price.create(
                unit_amount=amount,
                currency="usd",
                recurring={"interval": interval},
                product=productId,
            )
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_product_list(self, detalied: bool = True):
        """retrieve all producs"""
        if not detalied:
            try:
                return stripe.Product.list(active=True)
            except Exception as e:
                return {"message": str(e)}
        else:  # retrieve product price
            try:
                return stripe.Product.list()
            except Exception as e:
                return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_create(
        self,
        paymentId: str,
        name: str = "cristopher evangelista",
        email: str = "imurbatman12@gmail.com",
        description: str = "Myca subscriber",
    ):
        """create customer"""
        try:
            return stripe.Customer.create(
                name=name,
                email=email,
                description=description,
                payment_method=paymentId,
                invoice_settings={"default_payment_method": paymentId},
            )
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_get(self, customerId: str):
        """retrieve customer information"""
        try:
            return stripe.Customer.retrieve(customerId)
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_payment_add(self, paymentMethodId: str, customerId: str):
        """add customer payment method"""
        try:
            paymentMethods = self.stripe_retrieve_customer_payment_methods(customerId)

            paymentMethod = stripe.PaymentMethod.attach(
                paymentMethodId, customer=customerId
            )

            if len(paymentMethods.data) == 0:
                self.stripe_update_default_payment_method(customerId, paymentMethodId)

            paymentMethod.is_default = len(paymentMethods.data) == 0

            return paymentMethod

        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_payment_delete(self, paymentMethodId: str):
        """remove customer payment method"""
        try:
            return stripe.PaymentMethod.detach(paymentMethodId)
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_payment_get(self, customerId: str):
        """get customer list of payment methods"""
        try:
            return stripe.PaymentMethod.list(
                customer=customerId,
                type="card",
            )
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_payment_default(self, customerId: str, paymentMethodId: str):
        """update default payment method of customer"""
        try:
            setting = {"default_payment_method": paymentMethodId}
            return stripe.Customer.modify(customerId, invoice_settings=setting)
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_subscription_create(
        self, paymentId: str, name: str, email: str, priceId: str, customerId: str
    ):
        """
        create customer subscription
        TODO: name and email parameters not used!
        """
        try:
            # attach payment method to customer
            self.stripe_add_customer_payment_methods(paymentId, customerId)

            # set card to default payment method
            self.stripe_update_default_payment_method(customerId, paymentId)

            subscription = stripe.Subscription.create(
                customer=customerId,
                items=[
                    {"price": priceId},
                ],
                trial_period_days=30,
            )

            subscription.payment_method = paymentId

            return subscription
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_subscription_delete(self, subscriptionId: str):
        """cancel customer subscription"""
        try:
            return stripe.Subscription.delete(subscriptionId)
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_subscription_get(self, customerId: str):
        """retrieve customer subcription"""
        try:
            subscription = stripe.Subscription.list(customer=customerId)

            if len(subscription.data) == 0:
                return {"status": "inactive", "message": "Customer has no subscription"}

            subscription.pop("items", None)

            return subscription.data[0]
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_invoices_list(
        self, customerId: str, subscriptionId: str, limit: int = 10, lastItem: str = ""
    ):
        """retrieve customer list of invoices"""
        try:
            if lastItem != "":
                invoices = stripe.Invoice.list(
                    customer=customerId,
                    limit=limit,
                    starting_after=lastItem,
                    subscription=subscriptionId,
                )
            else:
                invoices = stripe.Invoice.list(
                    customer=customerId, limit=limit, subscription=subscriptionId
                )
            return invoices
        except Exception as e:
            return {"message": str(e)}



"""
Admin Global api functions as a mixin
"""
from jaseci.api.interface import interface
import jaseci.actions.live_actions as lact
import json


class actions_api:
    """
    Admin global APIs
    """

    @interface.admin_api(cli_args=["file"])
    def actions_load_local(self, file: str):
        """
        Hot load a python module and assimlate any Jaseci Actions
        """
        success = lact.load_local_actions(file)
        if success:
            cur_config = self.config_get("ACTION_SETS")
            if cur_config and (not isinstance(cur_config, list)):
                config = json.loads(cur_config)
                if file not in config["local"]:
                    config["local"].append(file)
                    self.config_set("ACTION_SETS", json.dumps(config))
            else:
                self.config_set(
                    "ACTION_SETS",
                    json.dumps({"local": [file], "remote": [], "module": []}),
                )
        return {"success": success}

    @interface.admin_api(cli_args=["url"])
    def actions_load_remote(self, url: str):
        """
        Hot load an actions set from live pod at URL
        """
        success = lact.load_remote_actions(url)
        if success:
            cur_config = self.config_get("ACTION_SETS")
            if cur_config and (not isinstance(cur_config, list)):
                config = json.loads(cur_config)
                if url not in config["remote"]:
                    config["remote"].append(url)
                    self.config_set("ACTION_SETS", json.dumps(config))
            else:
                self.config_set(
                    "ACTION_SETS",
                    json.dumps({"local": [], "remote": [url], "module": []}),
                )
        return {"success": success}

    @interface.admin_api(cli_args=["mod"])
    def actions_load_module(self, mod: str):
        """
        Hot load an actions set from live pod at URL
        """
        success = lact.load_module_actions(mod)
        if success:
            cur_config = self.config_get("ACTION_SETS")
            if cur_config and (not isinstance(cur_config, list)):
                config = json.loads(cur_config)
                if mod not in config["module"]:
                    config["module"].append(mod)
                    self.config_set("ACTION_SETS", json.dumps(config))
            else:
                self.config_set(
                    "ACTION_SETS",
                    json.dumps({"local": [], "remote": [], "module": [mod]}),
                )
        return {"success": success}

    # @interface.admin_api()
    # def actions_get(self, name: str, value: str):
    #     """
    #     """

    @interface.admin_api()
    def actions_list(self, name: str = ""):
        """
        List all live jaseci actions
        """
        actions = list(lact.live_actions.keys())
        if len(name):
            actions = list(filter(lambda a: a.startswith(name), actions))
        return actions

    # @interface.admin_api()
    # def actions_delete(self, name: str, value: str):
    #     """
    #     """



"""
Object api as a mixin
"""
from jaseci.api.interface import interface
from jaseci.element.element import element


class object_api:
    """Object APIs for generalized operations on Jaseci objects

    ...
    """

    def __init__(self):
        self.perm_default = "private"
        self._valid_perms = ["public", "private", "read_only"]

    @interface.private_api(cli_args=["name"])
    def global_get(self, name: str):
        """
        Get a global var
        """
        return {"value": self._h.get_glob(name)}

    @interface.private_api(cli_args=["obj"])
    def object_get(self, obj: element, depth: int = 0, detailed: bool = False):
        """Returns object details for any Jaseci object."""
        return obj.serialize(deep=depth, detailed=detailed)

    @interface.private_api(cli_args=["obj"])
    def object_perms_get(self, obj: element):
        """Returns object access mode for any Jaseci object."""
        return {"access": obj.j_access}

    @interface.private_api(cli_args=["obj"])
    def object_perms_set(self, obj: element, mode: str):
        """Sets object access mode for any Jaseci object."""
        ret = {}
        if mode not in self._valid_perms:
            ret["success"] = False
            ret["response"] = f"{mode} not valid, must be in {self._valid_perms}"
        else:
            getattr(obj, "make_" + mode)()
            ret["success"] = True
            ret["response"] = f"{obj} set to {mode}"
        return ret

    @interface.private_api(cli_args=["mode"])
    def object_perms_default(self, mode: str):
        """Sets object access mode for any Jaseci object."""
        ret = {}
        if mode not in self._valid_perms:
            ret["success"] = False
            ret["response"] = f"{mode} not valid, must be in {self._valid_perms}"
        else:
            self.perm_default = mode
            ret["success"] = True
            ret["response"] = f"Default access for future objects set to {mode}"
        return ret

    @interface.private_api(cli_args=["obj"])
    def object_perms_grant(self, obj: element, mast: element, read_only: bool = False):
        """Grants another user permissions to access a Jaseci object."""
        granted = obj.give_access(mast, read_only=read_only)
        ret = {"success": granted}
        if granted:
            ret["response"] = f"Access to {obj} granted to {mast}"
        else:
            ret["response"] = f"Cannot grant {mast} access to {obj}"
        return ret

    @interface.private_api(cli_args=["obj"])
    def object_perms_revoke(self, obj: element, mast: element):
        """Remove permissions for user to access a Jaseci object."""
        revoked = obj.remove_access(mast)
        ret = {"success": revoked}
        if revoked:
            ret["response"] = f"Access to {obj} revoked from {mast}"
        else:
            ret["response"] = f"{mast} did not have access to {obj}"
        return ret



"""
General action base class with automation for hot loading
"""
from jaseci.utils.utils import logger
from fastapi import FastAPI
from fastapi.responses import RedirectResponse
from pydantic import validate_arguments
from time import time
import inspect
import uvicorn
import os

remote_actions = {}
registered_apis = []
ACTIONS_SPEC_LOC = "/jaseci_actions_spec/"
JS_ACTION_PREAMBLE = "js_action_"


def mark_as_remote(api):
    registered_apis.append(api)


def serv_actions():
    """Returns fastAPI app interface for actions"""
    app = FastAPI(
        title="Jaseci Action Set API",
        description="A Jaseci Action Set",
    )

    @app.get("/")
    def home_redirect():
        return RedirectResponse(url="/docs")

    @app.get(ACTIONS_SPEC_LOC)
    def action_list():
        return remote_actions

    for i in registered_apis:
        gen_api_service(app, i[0], i[1], i[2], i[3])
    return app


def gen_api_service(app, func, act_group, aliases, caller_globals):
    """Helper for jaseci_action decorator"""
    # Construct list of action apis available
    varnames = list(inspect.signature(func).parameters.keys())
    act_group = (
        [os.path.basename(inspect.getfile(func)).split(".")[0]]
        if act_group is None
        else act_group
    )
    remote_actions[f"{'.'.join(act_group+[func.__name__])}"] = varnames

    # Need to get pydatic model for func signature for fastAPI post
    model = validate_arguments(func).model

    # Keep only feilds present in param list in base model
    keep_fields = {}
    for i in model.__fields__.keys():
        if i in varnames:
            keep_fields[i] = model.__fields__[i]
    model.__fields__ = keep_fields

    # Create duplicate funtion for api endpoint and inject in call site globals
    @app.post(f"/{func.__name__}/")
    def new_func(params: model):
        pl_peek = str(dict(params.__dict__))[:128]
        logger.info(str(f"Incoming call to {func.__name__} with {pl_peek}"))
        start_time = time()
        TY = "\033[33m"
        TG = "\033[32m"
        EC = "\033[m"  # noqa
        ret = validate_arguments(func)(**(params.__dict__))
        tot_time = time() - start_time
        logger.info(
            str(
                f"API call to {TG}{func.__name__}{EC}"
                f" completed in {TY}{tot_time:.3f} seconds{EC}"
            )
        )
        return ret

    for i in aliases:
        new_func = app.post(f"/{i}/")(new_func)
        remote_actions[f"{'.'.join(act_group+[i])}"] = varnames
    caller_globals[f"{JS_ACTION_PREAMBLE}{func.__name__}"] = new_func


def launch_server(port=80, host="0.0.0.0"):
    uvicorn.run(serv_actions(), port=port, host=host)



from unittest import TestCase
from jaseci.utils.utils import TestCaseHelper
from jaseci.jsctl import jsctl
from click.testing import CliRunner
import json


class jsctl_test(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()

    def call(self, cmd):
        res = CliRunner(mix_stderr=False).invoke(jsctl.jsctl, ["-m"] + cmd.split(" "))
        # self.log(res.stdout)
        # self.log(res.stderr)
        # self.log(res.exception)
        return res.stdout

    def call_cast(self, cmd):
        ret = self.call(cmd)
        # self.log(ret)
        return json.loads(ret)

    def call_split(self, cmd):
        ret = self.call(cmd)
        # self.log(ret)
        return ret.split("\n")

    def tearDown(self):
        jsctl.session["master"]._h.clear_mem_cache()
        super().tearDown()

    def test_jsctl_extract_tree(self):
        out = jsctl.extract_api_tree()
        self.assertIn("graph", out)
        self.assertIn("walker", out)
        self.assertIn("sentinel", out)
        self.assertIn("alias", out)
        self.assertIn("config", out)
        self.assertNotIn("jadc", out)

    def test_help_screen(self):
        r = self.call("--help")
        self.assertIn("Specify filename", r)
        self.assertIn("Group of `graph` commands", r)
        r = self.call("sentinel --help")
        self.assertIn("Group of `sentinel active` commands", r)

    def test_jsctl_create_graph_mem_only(self):
        self.call(
            "sentinel register " "jaseci/jsctl/tests/ll.jac -name ll -set_active true"
        )
        self.assertEqual(len(self.call_cast("graph list")), 1)
        self.call("graph create")
        self.call("graph create")
        self.call("graph create")
        self.assertEqual(len(self.call_cast("graph list")), 4)

    def test_jsctl_carry_forward(self):
        self.call("actions load local jaseci/tests/infer.py")
        self.call(
            "sentinel register " "jaseci/jsctl/tests/ll.jac -name ll -set_active true"
        )
        self.call("graph create -set_active true")
        self.call("walker run init")
        self.call("walker run gen_rand_life")
        r = self.call_cast("walker run get_gen_day")
        self.assertGreater(len(r["report"]), 3)

    def test_jsctl_dot(self):
        self.call(
            "sentinel register " "jaseci/jsctl/tests/ll.jac -name ll -set_active true"
        )
        self.call("graph create -set_active true")
        self.call("walker run init")
        self.call("walker run gen_rand_life")
        r = self.call("graph get -mode dot")
        self.assertIn('"n0" -> "n', r)
        self.assertIn('week="', r)

    def test_jsctl_aliases(self):
        """Tests that alias mapping api works"""
        self.call(
            "sentinel register " "jaseci/jsctl/tests/zsb.jac -name zsb -set_active true"
        )
        gph_id = self.call_cast("graph create")["jid"]
        snt_id = self.call_cast("sentinel list")[0]["jid"]
        self.call(f"alias register s -value {snt_id}")
        self.call(f"alias register g -value {gph_id}")
        self.assertEqual(len(self.call_cast("graph get -gph g")), 1)
        self.call("walker run init -snt s -nd g")
        self.assertEqual(len(self.call_cast("graph get -gph g")), 3)
        self.call("alias clear")
        self.assertEqual(len(self.call_cast("alias list").keys()), 0)

    def test_jsctl_auto_aliases(self):
        """Tests that auto alias mapping api works"""
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb " "-set_active true"
        )
        aliases = self.call_cast("alias list")
        self.assertGreater(len(aliases), 20)
        self.assertIn("zsb:architype:bot", aliases.keys())
        self.assertIn("zsb:walker:similar_questions", aliases.keys())
        self.assertIn("sentinel:zsb", aliases.keys())
        self.assertIn("zsb:walker:create_nugget", aliases.keys())

    def test_jsctl_auto_aliases_delete(self):
        """Tests that auto removing alias mapping api works"""
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb " "-set_active true"
        )
        num = len(self.call_cast("alias list"))

        self.call("architype delete zsb:architype:bot")
        self.call("walker delete zsb:walker:similar_questions")
        self.call("walker delete zsb:walker:create_nugget")
        self.assertEqual(num - 3, len(self.call_cast("alias list")))

    def test_jsctl_config_cmds(self):
        """Tests that config commands works"""
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb " "-set_active true"
        )
        self.call("config set CONFIG_EXAMPLE -value TEST -do_check False")
        self.call("config set APPLE -value Grape2 -do_check False")
        self.call('config set "Banana" -value "Grape" -do_check False')
        self.call('config set "Pear" -value "Banana" -do_check False')
        r = self.call("config get APPLE -do_check False")
        self.assertEqual(r.strip(), "Grape2")
        r = self.call_cast("config list")
        self.assertEqual(len(r), 1)

    def test_jsctl_default_snt_setting(self):
        """Tests that alias mapping api works"""
        self.call(
            "sentinel register jaseci/jsctl/tests/teststest.jac -name zsb "
            "-set_active true"
        )
        snt_id = self.call_cast("sentinel list")[0]["jid"]
        self.call(f"sentinel active set {snt_id}")
        self.call("sentinel active get")
        self.assertEqual(len(self.call_cast("walker list")), 2)

    def test_jsctl_init_auto_called(self):
        """Tests that alias mapping api works"""
        num = len(self.call_cast("graph get"))
        self.assertEqual(num, 3)

    def test_jsctl_master_defaults(self):
        """Tests that alias mapping api works"""
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb " "-set_active true"
        )
        gph_id = self.call_cast("graph create")["jid"]
        snt_id = self.call_cast("sentinel list")[0]["jid"]
        self.call(f"sentinel active set -snt {snt_id}")
        self.call(f"graph active set -gph {gph_id}")
        self.assertEqual(len(self.call_cast("graph get")), 1)
        self.call("walker run init")
        self.assertEqual(len(self.call_cast("graph get")), 3)

    def test_public_apis_present(self):
        r = self.call("walker --help")
        self.assertIn("summon", r)
        self.assertIn("namespace key", r)

    def test_public_apis_walker_summon_auth(self):
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb " "-set_active true"
        )
        r = self.call_cast("walker get zsb:walker:pubinit -mode keys")
        key = r["anyone"]
        r = self.call_cast("alias list")
        walk = r["zsb:walker:pubinit"]
        nd = r["active:graph"]
        r = self.call_cast(f"walker summon {walk} -key {key} -nd {nd}")
        self.assertEqual(len(r["report"]), 0)
        key = "aaaaaaaa"
        r = self.call_cast(f"walker summon {walk} -key {key} -nd {nd}")
        self.assertFalse(r["success"])

    def test_jsctl_import(self):
        r = self.call(
            "sentinel register "
            "jaseci/jsctl/tests/ll_base.jac -code_dir "
            "jaseci/jsctl/tests -set_active true"
        )
        self.call("walker run init")
        self.call("walker run gen_rand_life")
        r = self.call_cast("walker run get_gen_day")
        self.assertGreater(len(r["report"]), 3)

    def test_jsctl_import_filters(self):
        self.call("sentinel register " "jaseci/jsctl/tests/base.jac -set_active true")
        r = self.call_cast("walker run init")
        self.assertEqual(len(r["report"]), 8)

    def test_jsctl_import_filters1(self):
        self.call("sentinel register " "jaseci/jsctl/tests/base1.jac -set_active true")
        r = self.call_cast("walker run init")
        self.assertEqual(len(r["report"]), 8)

    def test_jsctl_import_filters2(self):
        self.call("sentinel register " "jaseci/jsctl/tests/base2.jac -set_active true")
        r = self.call_cast("walker run init")
        self.assertEqual(len(r["report"]), 8)

    def test_jsctl_import_fails_when_incomplete(self):
        self.call("sentinel register " "jaseci/jsctl/tests/base3.jac -set_active true")
        r = self.call_cast("walker run init")
        self.assertTrue("success" in r.keys())
        self.assertFalse(r["success"])

    def test_jsctl_import_globals(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/base4.jac -code_dir jaseci/jsctl/tests/ "
            "-set_active true"
        )
        r = self.call_cast("walker run init")
        self.assertEqual(len(r["report"]), 8)

    def test_jsctl_import_recursive(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/base5.jac -code_dir jaseci/jsctl/tests/ "
            "-set_active true"
        )
        r = self.call_cast("walker run init")
        self.assertEqual(r["report"][0], "plain")

    def test_jsctl_import_path(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/base6.jac -code_dir jaseci/jsctl/tests/ "
            "-set_active true"
        )
        r = self.call_cast("walker run init")
        self.assertEqual(r["report"][0], "plain")

    def test_jsctl_run_tests(self):
        self.call("sentinel register " "jaseci/jsctl/tests/teststest.jac")
        r = self.call_split("sentinel test")
        self.assertTrue(r[0].startswith('Testing "assert should be'))
        self.assertTrue(r[4].startswith('  "tests": 3'))
        self.assertTrue(r[7].startswith('  "success": true'))

    def test_jsctl_run_tests_detailed(self):
        self.call("sentinel register " "jaseci/jsctl/tests/teststest.jac")
        r = self.call_split("sentinel test -detailed true")
        self.assertEqual(len(r), 33)

    def test_jsctl_multiple_registers_with_globals(self):
        r = self.call_cast(
            "sentinel register " "jaseci/jsctl/tests/teststest.jir -mode ir"
        )
        r = self.call_cast(
            "sentinel register " "jaseci/jsctl/tests/teststest.jir -mode ir"
        )
        self.assertEqual(len(r), 1)

    def test_jsctl_run_tests_with_stdout(self):
        self.call("sentinel register " "jaseci/jsctl/tests/teststest_stdout.jac")
        r = self.call("sentinel test -detailed true")
        self.assertIn('"stdout": "Some Output\\nSome Output\\n"', r)
        r = r.replace('"stdout": "Some Output\\nSome Output\\n"', "")
        self.assertNotIn("Some Output", r)

    def test_jsctl_jac_build(self):
        import os

        if os.path.exists("jaseci/jsctl/tests/teststest.jir"):
            os.remove("jaseci/jsctl/tests/teststest.jir")
            self.assertFalse(os.path.exists("jaseci/jsctl/tests/teststest.jir"))
        self.call("jac build jaseci/jsctl/tests/teststest.jac")
        self.assertGreater(os.path.getsize("jaseci/jsctl/tests/teststest.jir"), 50000)

    def test_jsctl_jac_test(self):
        r = self.call_split("jac test jaseci/jsctl/tests/teststest.jac")
        self.assertTrue(r[0].startswith('Testing "assert should be'))
        self.assertTrue(r[4].startswith('  "tests": 3'))
        self.assertTrue(r[7].startswith('  "success": true'))

    def test_jsctl_jac_test_jir(self):
        r = self.call_split("jac test jaseci/jsctl/tests/teststest.jir")
        self.assertTrue(r[0].startswith('Testing "assert should be'))
        self.assertTrue(r[4].startswith('  "tests": 3'))
        self.assertTrue(r[7].startswith('  "success": true'))

    def test_jsctl_jac_run(self):
        r = self.call_cast("jac run jaseci/jsctl/tests/teststest.jac")
        self.assertEqual(r["report"], [{}, 4])

    def test_jsctl_jac_run_jir(self):
        r = self.call_cast("jac run jaseci/jsctl/tests/teststest.jir")
        self.assertEqual(r["report"], [{}, 4])

    def test_jsctl_jac_run_jir_walk(self):
        r = self.call_cast("jac run jaseci/jsctl/tests/teststest.jir -walk alt_init")
        self.assertEqual(r["report"], [7])

    def test_jsctl_sentinel_set(self):
        self.call(
            "sentinel register " "jaseci/jsctl/tests/teststest.jac -set_active true"
        )
        self.call("sentinel set " "jaseci/jsctl/tests/base.jac")
        r = self.call_cast("walker run init")
        r = self.call_cast("walker run init")
        self.assertEqual(len(r["report"]), 8)

    def test_jsctl_action_list(self):
        r = self.call_cast("actions list")
        self.assertIn("std.out", r)

    def test_jsctl_master_self(self):
        r = self.call_cast("master self")
        self.assertIn("j_type", r.keys())
        a = r["jid"]
        r = self.call_cast("master active get")
        self.assertIn("j_type", r.keys())
        b = r["jid"]
        self.assertEqual(a, b)

    def test_jsctl_graph_can(self):
        self.call("actions load local jaseci/tests/infer.py")
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/graph_can.jac -name gc -set_active true"
        )
        r = self.call("walker run go")
        self.assertEqual(r.split()[0], "2020-01-01T00:00:00")

    def test_jsctl_custom_report(self):
        self.call(
            "sentinel register " "jaseci/jsctl/tests/glob_imp.jac -set_active true"
        )
        r = self.call_cast("walker run cust_report")
        self.assertEqual(r, {"a": "b"})

    def test_jsctl_custom_report_off(self):
        self.call(
            "sentinel register " "jaseci/jsctl/tests/glob_imp.jac -set_active true"
        )
        r = self.call_cast("walker run cust_report_neutralize")
        self.assertIn("success", r.keys())

    def test_jsctl_disengage_report(self):
        self.call(
            "sentinel register " "jaseci/jsctl/tests/glob_imp.jac -set_active true"
        )
        r = self.call_cast("walker run disengage_report")
        self.assertEqual(r, {"a": "b"})

    def test_jsctl_set_global_default_perms(self):
        self.call("object perms default public")
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb " "-set_active true"
        )
        gphs = self.call_cast("graph get -detailed true")
        for i in gphs:
            self.assertEqual(i["j_access"], "public")



"""
Edge class for Jaseci

Each edge has an id, name, timestamp, the from node at the element of the edge
and the to node it is pointing to.
"""
from jaseci.element.element import element
from jaseci.element.obj_mixins import anchored
from jaseci.utils.utils import logger
import uuid


class edge(element, anchored):
    """Edge class for Jaseci"""

    def __init__(self, from_node=None, to_node=None, *args, **kwargs):
        self.from_node_id = None
        self.to_node_id = None
        self.bidirected: bool = False
        self.context = {}
        anchored.__init__(self)
        element.__init__(self, *args, **kwargs)
        if from_node:
            self.set_from_node(from_node)
        if to_node:
            self.set_to_node(to_node)

    def from_node(self):
        """Returns node edge is pointing from"""
        ret = (
            self._h.get_obj(self._m_id, uuid.UUID(self.from_node_id))
            if self.from_node_id
            else None
        )
        if not ret:
            logger.critical(str(f"{self} disconnected from source node"))
            return None
        else:
            return ret

    def to_node(self):
        """Returns node edge is pointing to"""
        if not self.to_node_id:
            return None
        ret = (
            self._h.get_obj(self._m_id, uuid.UUID(self.to_node_id))
            if self.to_node_id
            else None
        )
        if not ret:
            logger.critical(str(f"{self} disconnected to target node"))
            return None
        else:
            return ret

    def nodes(self):
        """Returns both nodes connected to edge in a list"""
        return [self.to_node(), self.from_node()]

    def opposing_node(self, node_obj):
        """Returns opposite node edge is pointing from node_obj"""
        node_set = [self.to_node_id, self.from_node_id]
        try:
            node_set.remove(node_obj.id.urn)
            return self._h.get_obj(self._m_id, uuid.UUID(node_set[0]))
        except ValueError:
            logger.critical(str(f"{self} disconnected to node {node_obj}"))
            return None

    def set_from_node(self, node_obj):
        """
        Returns node edge is pointing from
        TODO: should check prior nodes edge_ids if is a reset
        """
        if self.to_node_id:
            if not self.to_node().dimension_matches(node_obj, silent=False):
                return False
        self.from_node_id = node_obj.jid
        if self.jid not in node_obj.edge_ids:
            node_obj.edge_ids.add_obj(self)
        self.save()
        return True

    def set_to_node(self, node_obj):
        """
        Returns node edge is pointing to
        TODO: should check prior nodes edge_ids if is a reset
        """
        if self.from_node_id:
            if not self.from_node().dimension_matches(node_obj, silent=False):
                return False
        self.to_node_id = node_obj.jid
        if self.jid not in node_obj.edge_ids:
            node_obj.edge_ids.add_obj(self)
        self.save()
        return True

    def set_bidirected(self, bidirected: bool):
        """Sets/unsets edge to be bidirected"""
        self.bidirected = bidirected
        self.save()

    def is_bidirected(self):
        """Check if edge is bidirected"""
        return self.bidirected

    def connects(self, source=None, target=None, ignore_direction=False):
        """Test if a node or nodes are connected by edge"""
        if not source and not target:
            return False
        if self.bidirected or ignore_direction:
            if source and source.id.urn not in [self.from_node_id, self.to_node_id]:
                return False
            if target and target.id.urn not in [self.from_node_id, self.to_node_id]:
                return False
        else:
            if source and source.id.urn != self.from_node_id:
                return False
            if target and target.id.urn != self.to_node_id:
                return False
        return True

    def set_context(self, ctx, arch=None):
        """Assign values to context of edge"""
        if arch is None:
            arch = self
        for i in ctx.keys():
            if i not in arch.context.keys():
                logger.warning(str(f"{i} not a context member of {self}"))
                continue
            else:
                self.context[i] = ctx[i]
        self.save()

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        base = self.from_node()
        target = self.to_node()
        if base and self.jid in base.edge_ids:
            base.edge_ids.remove_obj(self)
        if target and self.jid in target.edge_ids:
            target.edge_ids.remove_obj(self)
        element.destroy(self)

    def dot_str(self, node_map=None, edge_map=None):
        """
        DOT representation
        from_node -> to_node [context_key=contect_value]
        """
        from_name = (
            uuid.UUID(self.from_node().jid).hex
            if node_map is None
            else node_map.index(self.from_node().jid)
        )
        to_name = (
            uuid.UUID(self.to_node().jid).hex
            if node_map is None
            else node_map.index(self.to_node().jid)
        )
        dstr = f'"n{from_name}" -> "n{to_name}" '

        dstr += f'[ id="{uuid.UUID(self.jid).hex}"'
        label = ""
        if edge_map:
            label = f"e{edge_map.index(self.jid)}"
        if self.name != "generic":
            label += f":{self.name}"
        if label:
            dstr += f', label="{label}"'
        if self.bidirected:
            dstr += ', dir="both"'

        edge_dict = self.context

        if edge_dict:
            for k, v in edge_dict.items():
                if not isinstance(v, str) or v == "":
                    continue
                dstr += f', {k}="{v[:32]}"'

        dstr += " ]"

        return dstr + "\n"



"""Built in actions for Jaseci"""
import requests
from jaseci.actions.live_actions import jaseci_action
from base64 import b64decode, b64encode
from io import BytesIO


@jaseci_action()
def get(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.get(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def post(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.post(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def put(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.put(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def delete(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.delete(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def head(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.head(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def options(url: str, data: dict, header: dict):
    """
    Issue request
    Param 1 - url
    Param 2 - data
    Param 3 - header

    Return - response object
    """
    res = requests.options(url, json=data, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def multipart_base64(url: str, files: list, header: dict):
    """
    Issue request
    Param 1 - url
    Param 3 - header
    Param 3 - file (Optional) used for single file
    Param 4 - files (Optional) used for multiple files
    Note - file and files can't be None at the same time

    Return - response object
    """

    if not files:
        return {
            "status_code": 400,
            "error": "Please include base64 using this format "
            '{"field":val,"name":val,"base64":val} '
            "using parameter `file` and `files` for array file",
        }

    formData = []

    if files is not None:
        for f in files:
            formData.append(
                (
                    f["field"] if "field" in f else "file",
                    (f["name"], BytesIO(b64decode(f["base64"]))),
                )
            )

    res = requests.post(url, files=formData, headers=header)
    ret = {"status_code": res.status_code}
    try:
        ret["response"] = res.json()
    except Exception:
        ret["response"] = res.text
    return ret


@jaseci_action()
def file_download_base64(url: str, header: dict, encoding: str = "utf-8"):
    """Standard built in for download file from url"""
    with requests.get(url, stream=True, headers=header) as res:
        res.raise_for_status()
        with BytesIO() as buffer:
            for chunk in res.iter_content(chunk_size=8192):
                buffer.write(chunk)
            ret = buffer.getvalue()
    return b64encode(ret).decode(encoding)



"""
The Jaseci Jac virtual machine for jsx bytecode
"""

from jaseci.jac.machine.machine_state import machine_state
from jaseci.jac.machine.jac_scope import jac_scope


class jsx_machine(machine_state):
    """Create 'bit switch' for jsx bytecode"""

    def __init__(self, jsx):
        self.PC = 0
        self.jsx = jsx

    def IDS_CLEAR(self, *args):
        """IDS_CLEAR op code"""
        self.get_ref(args[0]).remove_all()

    def IDS_ADD_OBJ(self, *args):
        """IDS_ADD_OBJ op code"""
        self.get_ref(args[0]).add_obj(self.get_ref(args[1]))

    def IDS_GET_LEN(self, *args):
        """IDS_GET_LEN op code"""
        self.set_ref(args[0], len(self.get_ref(args[1])))

    def PUSH_SCOPE_W(self, *args):
        """PUSH_SCOPE_W op code"""
        self.push_scope(
            jac_scope(
                parent=self,
                has_obj=self,
                action_sets=[
                    self.activity_action_ids,
                    self.current_node.activity_action_ids,
                ],
            )
        )

    def POP_SCOPE(self, *args):
        """POP_SCOPE op code"""
        self.pop_scope()

    def SET_LIVE_VAR(self, *args):
        """SET_LIVE_VAR op code"""
        self._jac_scope.set_live_var(
            name=args[0], value=self.get_ref(args[1]), md_index=args[2], jac_ast=None
        )

    def SET_REF_VAR(self, *args):
        """SET_REF_VAR op code"""
        self.set_ref(args[0], self.get_ref(args[1]))

    def SET_REF_VARI(self, *args):
        """SET_REF_VARI op code"""
        self.set_ref(args[0], args[1])

    def CREATE_CTX_VAR(self, *args):
        """CREATE_CTX_VAR op code"""
        obj = self.get_ref(args[0])
        var_name = args[1]
        var_val = self.get_ref(args[2])
        is_private = args[3]
        if var_name not in obj.context.keys():  # Runs has once per walk
            obj.context[var_name] = var_val
        if is_private:
            if "_private" in obj.context.keys():
                if var_name not in obj.context["_private"]:
                    obj.context["_private"].append(var_name)
            else:
                obj.context["_private"] = [var_name]

    def SET_ANCHOR(self, *args):
        """SET_ANCHOR op code"""
        obj = args[0]
        var_name = args[1]
        if "anchor" in dir(obj):
            if obj.anchor is None:
                obj.anchor = var_name

    def B_NEQ(self, *args):
        """B_NEQ op code"""
        if self.get_ref(args[0]) != self.get_ref(args[1]):
            self.PC += args[2]

    def B_NEQI(self, *args):
        """B_NEQI op code"""
        if self.get_ref(args[0]) != args[1]:
            self.PC += args[2]

    def B_NIT(self, *args):
        """B_NIT op code"""
        if not isinstance(self.get_ref(args[0]), args[1]):
            self.PC += args[2]

    def B_A(self, *args):
        """B_NEQ op code"""
        self.PC += args[0]

    def PLUS(self, *args):
        """B_NEQ op code"""
        self.set_ref(
            self.get_ref(args[0]), self.get_ref(args[1]) + self.get_ref(args[2])
        )

    def END(self, *args):
        """END op code"""
        self.PC = len(self.jsx)

    def get_ref(self, r):
        """Access ref from instruction"""

    def set_ref(self, r, val):
        """Access ref from instruction"""



"""
ID list class for Jaseci

Generalized functions for managing '_ids' convention for lists of Jaseci
objects

parent_obj is the instance that the list belongs to
"""
from jaseci.utils.utils import logger
import uuid


class id_list(list):
    """
    ID list class for tracking lists of objects in Jaseci

    ingest_list is a list of hex strings to convert to UUID and append.
    """

    def __init__(self, parent_obj, in_list=None):
        self.parent_obj = parent_obj
        if in_list:
            for i in in_list:
                self.append(i)

    def add_obj(self, obj, allow_dups=False, silent=False):
        """Adds a obj obj to Jaseci object"""
        self.parent_obj.check_hooks_match(obj)
        if not allow_dups and obj.jid in self:
            if not silent:
                logger.warning(str(f"{obj} is already in {self.parent_obj}'s list"))
        else:
            self.append(obj.jid)
            if not obj.parent_id:
                obj.parent_id = self.parent_obj.id
            obj.save()
            self.parent_obj.save()

    def add_obj_list(self, obj_list, allow_dups=False, silent=False):
        for i in obj_list:
            self.add_obj(i, allow_dups=allow_dups, silent=silent)

    def remove_obj(self, obj):
        """Remove a Jaseci obj from list"""
        self.remove(obj.jid)
        self.parent_obj.save()

    def destroy_obj(self, obj):
        """Completely destroys a Jaseci obj obj by it's name"""
        self.remove_obj(obj)
        obj.destroy()

    def get_obj_by_name(self, name, kind=None, silent=False):
        """Returns a Jaseci obj obj by it's name"""
        healing = []
        ret = None
        for i in self:
            if not self.parent_obj._h.has_obj(uuid.UUID(i)):
                logger.critical(
                    str(
                        f"Self healing: {i} not found "
                        + f"in id_list of {self.parent_obj}!"
                    )
                )
                healing.append(i)
                continue
            obj = self.parent_obj._h.get_obj(self.parent_obj._m_id, uuid.UUID(i))
            if obj.name == name:
                if kind and obj.kind != kind:
                    continue
                ret = self.parent_obj._h.get_obj(self.parent_obj._m_id, uuid.UUID(i))
                break
        self.heal(healing)
        if not ret and not silent:
            logger.error(str(f"object for '{name}' not found in '{self.parent_obj}'!"))
        return ret

    def has_obj_by_name(self, name, kind=None):
        """Returns whether a Jaseci obj exists by it's name"""
        return self.get_obj_by_name(name, kind, silent=True) is not None

    def remove_obj_by_name(self, name, kind=None):
        """Remove a Jaseci obj by it's name"""
        self.remove_obj(self.get_obj_by_name(name, kind))

    def destroy_obj_by_name(self, name, kind=None):
        """Destroy a Jaseci obj by it's name"""
        self.destroy_obj(self.get_obj_by_name(name, kind))

    def obj_list(self):
        """Return list of objects from ids"""
        ret = []
        healing = []
        for i in self:
            obj = self.parent_obj._h.get_obj(self.parent_obj._m_id, uuid.UUID(i))
            if not obj:
                logger.critical(
                    str(
                        f"Self healing: {i} not found "
                        + f"in id_list of {self.parent_obj}!"
                    )
                )
                healing.append(i)
            else:
                ret.append(
                    self.parent_obj._h.get_obj(self.parent_obj._m_id, uuid.UUID(i))
                )
        self.heal(healing)
        return ret

    def remove_all(self):
        """Remove a Jaseci obj obj by it's name"""
        for i in self.obj_list():
            self.remove_obj(i)
        if len(self):
            logger.critical(
                str(
                    f"Removeall all failed in id_list of {self.parent_obj} - "
                    + f"still has {self}!"
                )
            )

    def destroy_all(self):
        """Remove a Jaseci obj obj by it's name"""
        for i in self.obj_list():
            self.destroy_obj(i)
        if len(self):
            logger.critical(
                str(
                    f"Destroy all failed in id_list of {self.parent_obj} - "
                    + f"still has {self}!"
                )
            )

    def heal(self, healing):
        if len(healing):
            for i in healing:
                self.remove(i)
            self.parent_obj.save()

    def first_obj(self):
        """Get first object in list"""
        if not self:
            logger.error(str(f"List in '{self.parent_obj}' is empty!"))
            return None
        return self.parent_obj._h.get_obj(self.parent_obj._m_id, uuid.UUID(self[0]))

    def pop_first_obj(self):
        """Get first object in list"""
        ret = self.first_obj()
        if ret:
            self.remove_obj(ret)
        return ret



"""
Jaseci object mixins

Various mixins to define properties of Jaseci objects
"""
from jaseci.utils.id_list import id_list
from jaseci.utils.utils import logger
import uuid


class anchored:
    """Utility class for objects that hold anchor values"""

    def __init__(self):
        self.anchor = None

    def anchor_value(self):
        """Returns value of anchor context object"""
        if self.anchor:
            return self.context[self.anchor]
        return None


class sharable:
    """Utility class for objects that are sharable between users"""

    def __init__(self, m_id, mode=None):
        self.set_master(m_id)
        self.j_access = (
            mode
            if mode is not None
            else self._h.get_obj(self._m_id, uuid.UUID(self._m_id)).perm_default
            if self._h.get_obj(self._m_id, uuid.UUID(self._m_id))
            else "private"
        )
        self.j_r_acc_ids = id_list(self)
        self.j_rw_acc_ids = id_list(self)

    @property
    def _m_id(self) -> str:
        return self.j_master

    def set_master(self, m_id):
        if m_id is None or m_id == "anon":
            m_id = uuid.UUID(int=0).urn
        self.j_master = m_id

    def make_public(self):
        """Make element publically accessible"""
        self.j_access = "public"
        self.save()

    def make_read_only(self):
        """Make element publically readable"""
        self.j_access = "read_only"
        self.save()

    def make_private(self):
        """Make element private"""
        self.j_access = "private"
        self.save()

    def is_public(self):
        """Check if element is publically accessible"""
        return self.j_access == "public" or self.j_master == uuid.UUID(int=0).urn

    def is_read_only(self):
        """Check if element is publically readable"""
        return self.j_access == "read_only"

    def is_readable(self):
        """Check if element is publically readable"""
        return self.j_access == "read_only" or self.is_public()

    def is_private(self):
        """Check if element is private"""
        return self.j_access == "private"

    def super_check(self, caller_id):
        """Quick check if caller is super master"""
        if not hasattr(self, "_h"):
            return False
        user = self._h.get_obj(caller_id, uuid.UUID(caller_id), override=True)
        if user.j_type == "super_master":
            return True
        return False

    def check_read_access(self, caller_id, silent=False):
        if (
            caller_id == self._m_id
            or self.is_readable()
            or caller_id in self.j_r_acc_ids
            or caller_id in self.j_rw_acc_ids
            or self.super_check(caller_id)
        ):
            return True
        if not silent:
            logger.error(str(f"{caller_id} does not have permission to access {self}"))
        return False

    def check_write_access(self, caller_id, silent=False):
        if (
            caller_id == self._m_id
            or self.is_public()
            or caller_id in self.j_rw_acc_ids
            or self.super_check(caller_id)
        ):
            return True
        if not silent:
            logger.error(str(f"{caller_id} does not have permission to access {self}"))
        return False

    def give_access(self, m, read_only=True):
        """Give access to a master (user)"""
        if not m.is_master():
            logger.error(f"{m} is not master!")
            return False
        if read_only and m.jid not in self.j_r_acc_ids:
            self.j_r_acc_ids.add_obj(m)
        elif m.jid not in self.j_rw_acc_ids:
            self.j_rw_acc_ids.add_obj(m)
        return True

    def remove_access(self, m):
        """Remove access from a master (user)"""
        ret = False
        if m.jid in self.j_r_acc_ids:
            self.j_r_acc_ids.remove_obj(m)
            ret = True
        if m.jid in self.j_rw_acc_ids:
            self.j_rw_acc_ids.remove_obj(m)
            ret = True
        return ret


class hookable(sharable):
    """Utility class for objects that are savable to DBs and other stores"""

    def __init__(self, h, m_id, persist: bool = True, *args, **kwargs):
        self._h = h  # hook for storing and loading to persistent store
        self._persist = persist
        sharable.__init__(self, m_id, *args, **kwargs)

    def check_hooks_match(self, target, silent=False):
        """Checks whether target object hook matches self's hook"""
        if not silent and target._h != self._h:
            logger.critical(
                str(
                    "Hook for {} does not match {}, {} != {}".format(
                        target, self, target._h, self._h
                    )
                )
            )
        return target._h == self._h

    def save(self):
        """
        Write self through hook to persistent storage
        """
        self._h.save_obj(self._m_id, self, self._persist)

    def destroy(self):
        """
        Destroys self from persistent storage

        Note that the object will still exist in python until GC'd
        """
        self._h.destroy_obj(self._m_id, self, self._persist)
        del self

    def parent(self):
        """
        Returns the objects for list of owners of this element
        """
        return self._h.get_obj(self._m_id, self.parent_id)



def load_standard():
    import jaseci.actions.standard.net  # noqa
    import jaseci.actions.standard.rand  # noqa
    import jaseci.actions.standard.request  # noqa
    import jaseci.actions.standard.std  # noqa
    import jaseci.actions.standard.file  # noqa
    import jaseci.actions.standard.vector  # noqa
    import jaseci.actions.standard.date  # noqa
    import jaseci.actions.standard.jaseci  # noqa


load_standard()



"""
Walker class for Jaseci

Each walker has an id, name, timestamp and it's parent sentinel.
TODO: Perhaps  I should have walker state (context ids) in mem only with
default hooks to save db read/writes
"""

from jaseci.utils.utils import logger
from jaseci.element.element import element
from jaseci.element.obj_mixins import anchored
from jaseci.utils.id_list import id_list
from jaseci.jac.interpreter.walker_interp import walker_interp
from jaseci.jac.ir.jac_code import jac_code
import uuid
import hashlib
import io
import pstats
import cProfile


class walker(element, jac_code, walker_interp, anchored):
    """Walker class for Jaseci"""

    def __init__(self, code_ir=None, *args, **kwargs):
        self.activity_action_ids = id_list(self)
        self.namespaces = []
        self.context = {}
        self.profile = {}
        # Process state
        self.current_node_id = None
        self.next_node_ids = id_list(self)
        self.ignore_node_ids = id_list(self)
        self.destroy_node_ids = id_list(self)
        self.current_step = 0
        self.in_entry_exit = False
        self.step_limit = 10000
        anchored.__init__(self)
        element.__init__(self, *args, **kwargs)
        jac_code.__init__(self, code_ir=code_ir)
        walker_interp.__init__(self)

    @property
    def current_node(self):
        if not self.current_node_id:
            return None
        else:
            return self._h.get_obj(self._m_id, uuid.UUID(self.current_node_id))

    @current_node.setter
    def current_node(self, obj):
        if obj:
            self.current_node_id = obj.id.urn
        else:
            self.current_node_id = None

    def namespace_keys(self):
        """Return list of md5 keys for namespaces"""
        ret = {}
        for i in self.namespaces:
            ret[i] = hashlib.md5((self._m_id + i).encode()).hexdigest()
        return ret

    def step(self):
        """
        Take single step through program
        if no ast provided, will be generated from code
        """
        if not self.next_node_ids:
            logger.debug(str(f"Walker {self.name} is disengaged"))
            return False
        if self.current_step > self.step_limit:
            logger.error(
                str(
                    f"Walker {self.name} walked too many steps "
                    + f"- {self.step_limit}"
                )
            )
            return False

        self.current_node = self.next_node_ids.pop_first_obj()
        self.run_walker(jac_ast=self._jac_ast)
        if self.current_step < 200:
            self.log_history("visited", self.current_node.id)
        self.current_step += 1
        self.profile["steps"] = self.current_step
        if self._stopped == "skip":
            self._stopped = None
        if self.next_node_ids:
            logger.debug(
                str(
                    f"Step complete, Walker {self.name} next node: "
                    + f"- {self.next_node_ids.first_obj()}"
                )
            )
            return self.next_node_ids.first_obj()
        else:
            logger.debug(
                str(
                    f"Final step of walker {self.name} complete"
                    + f"- disengaging from {self.current_node}"
                )
            )
            for i in self.destroy_node_ids.obj_list():
                i.destroy()
            return True

    def prime(self, start_node, prime_ctx=None, request_ctx=None):
        """Place walker on node and get ready to step step"""
        self.clear_state()
        self.next_node_ids.add_obj(start_node)
        if prime_ctx:
            for i in prime_ctx.keys():
                self.context[str(i)] = prime_ctx[i]
        self.request_context = request_ctx
        self.profile["steps"] = self.current_step
        logger.debug(str(f"Walker {self.name} primed - {start_node}"))

    def run(self, start_node=None, prime_ctx=None, request_ctx=None, profiling=False):
        """Executes Walker to completion"""
        if profiling:
            pr = cProfile.Profile()
            pr.enable()

        if start_node:
            self.prime(start_node, prime_ctx, request_ctx)

        report_ret = {"success": True}
        try:
            while self.step():
                pass
        except Exception as e:
            import traceback as tb

            self.rt_error(f"Internal Exception: {e}", self._cur_jac_ast)
            report_ret["stack_trace"] = tb.format_exc()

        self.save()

        if not self.report:
            logger.debug(str(f"Walker {self.name} did not arrive at report state"))

        report_ret["report"] = self.report

        if self.report_status:
            report_ret["status_code"] = self.report_status
        if self.report_custom:
            report_ret["report_custom"] = self.report_custom
        if len(self.runtime_errors):
            report_ret["errors"] = self.runtime_errors
            report_ret["success"] = False
        if profiling:
            pr.disable()
            s = io.StringIO()
            sortby = pstats.SortKey.CUMULATIVE
            ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
            ps.print_stats()
            s = s.getvalue()
            s = "ncalls" + s.split("ncalls")[-1]
            s = "\n".join(
                [",".join(line.rstrip().split(None, 5)) for line in s.split("\n")]
            )
            self.profile["perf"] = s
            report_ret["profile"] = self.profile

        return report_ret

    def log_history(self, name, value):
        """Helper function for logging history of walker's activities"""
        if isinstance(value, element):
            value = {"type": value.j_type, "id": value.id.urn}
        if isinstance(value, uuid.UUID):
            value = value.urn
        if name in self.profile.keys():
            self.profile[name].append(value)
        else:
            self.profile[name] = [value]

    def clear_state(self):
        """Clears walker state after report"""
        self.profile = {}
        self.current_step = 0
        self.next_node_ids.remove_all()
        self.ignore_node_ids.remove_all()
        self.destroy_node_ids.remove_all()
        self.current_node = None
        self.activity_action_ids.destroy_all()
        self.context = {}
        walker_interp.reset(self)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.activity_action_ids.obj_list():
            i.destroy()
        super().destroy()



"""
Logger api as a mixin
"""
from jaseci.api.interface import interface
from jaseci.utils.utils import logger, app_logger
from jaseci.utils.utils import connect_logger_handler
from logging.handlers import HTTPHandler


class logger_api:
    """
    APIs for Jaseci Logging configuration
    """

    @interface.admin_api(cli_args=["host"])
    def logger_http_connect(self, host: str, port: int, url: str, log: str = "all"):
        """
        Connects internal logging to http(s) (log msgs sent via POSTs)
        Valid log params: {sys, app, all }
        """
        num = 0
        if log == "sys" or log == "all":
            connect_logger_handler(
                logger, HTTPHandler(host=f"{host}:{port}", url=url, method="POST")
            )
            num += 1
        if log == "app" or log == "all":
            connect_logger_handler(
                app_logger, HTTPHandler(host=f"{host}:{port}", url=url, method="POST")
            )
            num += 1
        return [f"{num} http handlers added!"]

    @interface.admin_api()
    def logger_http_clear(self, log: str = "all"):
        """
        Connects internal logging to http(s) (log msgs sent via POSTs)
        Valid log params: {sys, app, all }
        """
        num = 0
        if log == "sys" or log == "all":
            for i in logger.handlers:
                if i.__class__.__name__ == "HTTPHandler":
                    logger.removeHandler(i)
                    num += 1
        if log == "app" or log == "all":
            for i in app_logger.handlers:
                if i.__class__.__name__ == "HTTPHandler":
                    app_logger.removeHandler(i)
                    num += 1
        return [f"{num} http handlers removed!"]

    @interface.admin_api()
    def logger_list(self):
        """
        Check active loggers
        """
        core = []
        app = []
        for i in logger.handlers:
            core.append(str(type(i)))
        for i in app_logger.handlers:
            app.append(str(type(i)))
        return {"core": core, "app": app}



"""Built in actions for Jaseci"""
from jaseci.actions.live_actions import jaseci_action
from jaseci.utils.utils import master_from_meta
from jaseci.jac.jac_set import jac_set
import uuid


@jaseci_action()
def max(item_set: jac_set):
    ret = None
    if not len(item_set):
        return None
    items = item_set.obj_list()
    max_val = items[0].anchor_value()
    ret = items[0]
    for i in items:
        if i.anchor_value() > max_val:
            ret = i
            max_val = i.anchor_value()
    return ret


@jaseci_action()
def min(item_set: jac_set):
    ret = None
    if not len(item_set):
        return None
    items = item_set.obj_list()
    min_val = items[0].anchor_value()
    ret = items[0]
    for i in items:
        if i.anchor_value() < min_val:
            ret = i
            min_val = i.anchor_value()
    return ret


@jaseci_action()
def root(meta):
    mast = master_from_meta(meta)
    if mast.active_gph_id:
        return mast._h.get_obj(mast._m_id, uuid.UUID(mast.active_gph_id))
    return None
